package francois.pessaux.androjig.android;

import java.io.File ;
import java.io.FileInputStream ;

import android.graphics.Bitmap ;
import android.graphics.BitmapFactory ;
import android.graphics.Matrix ;
import francois.pessaux.androjig.android.GameStruct.Game ;
import francois.pessaux.androjig.android.GameStruct.Layout_cell ;
import francois.pessaux.androjig.android.Maths2d.Point ;
import francois.pessaux.androjig.android.PieceStruct.Connected_set ;
import francois.pessaux.androjig.android.PieceStruct.Edge ;
import francois.pessaux.androjig.android.PieceStruct.Piece ;
import francois.pessaux.androjig.android.PieceStruct.Pieces_list ;
import francois.pessaux.androjig.android.SplineStruct.Spline ;

public class GameAlloc {

  /** Insert [edge] after the edge [after] in the double linked list of edges
      sorted by increasing minimal x coordinate. This function is used
      at list initialization-time. It doesn't check the coordinates, so the
      fact that edges as they are inserted are well sorted is uniquely ensured
      by the algorithm that calls this function. */
  private static void insertAfterInXList (Edge edge, Edge after)
  {
    edge.x_sorted_next = after.x_sorted_next ;
    after.x_sorted_next = edge ;
    if (edge.x_sorted_next != null)
      edge.x_sorted_next.x_sorted_prev = edge ;
      edge.x_sorted_prev = after ;
  }



  /** Insert [edge] after the edge [after] in the double linked list of edges
      sorted by increasing minimal y coordinate. This function is used
      at list initialization-time. It doesn't check the coordinates, so the
      fact that edges as they are inserted are well sorted is uniquely ensured
      by the algorithm that calls this function. */
  private static void InsertAfterInYList (Edge edge, Edge after)
  {
    edge.y_sorted_next = after.y_sorted_next ;
    after.y_sorted_next = edge ;
    if (edge.y_sorted_next != null)
      edge.y_sorted_next.y_sorted_prev = edge ;
      edge.y_sorted_prev = after ;
  }



  private static Bitmap makeScaledCopyOfImageAndAddAlphaChannel (Bitmap image,
      float scale_factor)
  {
    /* Result image. */
    Matrix m = new Matrix () ;
    m.preScale (scale_factor, scale_factor) ;
    Bitmap scaled_image =
      Bitmap.createBitmap
        (image, 0, 0, image.getWidth (), image.getHeight (), m, true) ;
    Bitmap scaled_with_alpha_bitmap =
      scaled_image.copy (Bitmap.Config.ARGB_8888, false) ;
    return (scaled_with_alpha_bitmap) ;
  }



  private static Bitmap copyImageAndAddAlphaChannel (Bitmap image)
  {
    Bitmap image_with_alpha = image.copy (Bitmap.Config.ARGB_8888, false) ;   
    return (image_with_alpha) ;
  }


  private static Bitmap rotateBitmap (Bitmap image)
  {
    Matrix m = new Matrix () ;
    m.preRotate ((float) 90.0) ;
    Bitmap rotatedImage =
      Bitmap.createBitmap
        (image, 0, 0, image.getWidth (), image.getHeight (), m, true) ;
    return (rotatedImage) ;
  }
  

  /** Loads an image from a file name and returns by side effect its width and
     height as a Point (if it is not null) where x is width and y is height. */
  public static Bitmap loadGameImage (String image_file_name,
                                  float scale_factor, Point width_n_height_out,
                                  boolean rotate)
  {
    
    Bitmap bitmap = BitmapFactory.decodeFile (image_file_name) ;
    if (bitmap == null) {
      if (width_n_height_out != null) {
        width_n_height_out.x = 0 ;
        width_n_height_out.y = 0 ;
      }
      return (null) ;
    }
   
    /* Modify the image size according to the scaling factor if needed and
       either scale the image and add transparency or simply check if we need to
       add transparency. */
    if ((scale_factor > 0) && (scale_factor != (float) 1.0)) {
      /* No need to check if we need to add transparency, the scaling routine
        always add it by the way because this doesn't cost it more. */
      bitmap = makeScaledCopyOfImageAndAddAlphaChannel (bitmap, scale_factor) ;
    }
    else {
      /* If the image doesn't have any transparency channel, add one. */
      if (! bitmap.hasAlpha ()) bitmap = copyImageAndAddAlphaChannel (bitmap) ;
    }

    /* If rotation is requested then really rotate. But do this only after
       transparency and scaling are done. */
    if (rotate) bitmap = rotateBitmap (bitmap) ;
    if (width_n_height_out != null) {
      width_n_height_out.x = bitmap.getWidth () ;
      width_n_height_out.y = bitmap.getHeight () ;
    }
    return (bitmap) ;
  }



  private static byte[] rawLoadImageFile (String image_file_name)
  {
    int file_size ;
    byte [] data ;
    File in = new File (image_file_name) ;
    FileInputStream fis = null ;
    try {
      fis = new FileInputStream (in) ;
      file_size = fis.available () ;
      data = new byte [file_size] ;
      fis.read (data, 0, file_size) ;
      fis.close () ;
      return (data) ;
    }
    catch (Exception e) {
      return (null) ;
    }
  }



   /** Initiates a new game creation.
       Note that if scale_factor is negative, null or 1.0, then this means that
       no scaling must be applied to the image. */
  public static Game newGame (String image_filename, int screen_width,
                              int screen_height, int nb_columns_of_pieces,
                              int nb_lines_of_pieces, float scale_factor,
                              boolean rotate)
  {
    int x, y, i ;
    int piece_width, piece_height ;
    int layout_nb_columns, layout_nb_lines ;
    Point image_dimensions = new Point () ;
    int image_width, image_height ;
    /* Matrix representing the pieces of the jigsaw in term of column/line
       position. We allocate this matrix as a contiguous array and index it by
       (y * nb_columns) + x. This array contains the pieces whose
       address will be shared everywhere. */
    PieceStruct.Piece[] pieces = null ;
    /* Matrix containing pieces that are (even partially) in the cell
       corresponding to the indices of this cell. The layout has many
       rows and columns the jigsaw has currently but could be as large
       as the playfield is. */
    GameStruct.Layout_cell[] layout = null ;
    /* Temporary edge pointer for various usages. */
    Edge tmp_edge ;
    /* Pointer used to initialize the sorted double linked lists. */
    Edge prev ;
    Bitmap game_image ;
    Integer raw_image_file_size = 0 ;
    byte[] raw_image_file_data = null ;
    /* The final game structure we will return. */
    Game game = null ;

    game_image =
      loadGameImage (image_filename, scale_factor, image_dimensions, rotate) ;
    if (game_image == null) return (null) ;
    
    /* Recover width and height of the image from the Point struct passed as
       out mode parameter to loadGameImage. */
    image_width = image_dimensions.x ;
    image_height = image_dimensions.y ;
    
    raw_image_file_data = rawLoadImageFile (image_filename) ;
    if (raw_image_file_data == null) return (null) ;

    raw_image_file_size = raw_image_file_data.length ;
    
    /* Compute the pieces width and height according to the number of
       columns and lines of pieces requested and the size of the
       image. */
    piece_width = image_width / nb_columns_of_pieces ;
    piece_height = image_height / nb_lines_of_pieces ;

    /* Attention, if the image size is not exactly divisible by the number of
       pieces, we must crop the image size to make so that the number of pieces
       times the pieces size is exactly the image size. */
    image_width = nb_columns_of_pieces * piece_width ;
    image_height = nb_lines_of_pieces * piece_height ;

    /* Allocate all the pieces of the jigsaw, initially seen as a matrix
       of size [nb_columns_of_pieces * nb_lines_of_pieces]. */
    pieces = new PieceStruct.Piece [nb_columns_of_pieces * nb_lines_of_pieces] ;
   
    for (i = 0; i < nb_columns_of_pieces * nb_lines_of_pieces; i++) {
      Piece piece = new Piece () ;
      pieces[i] = piece ;
      piece.piece_debug_id  = 0 ;
      piece.center = null ;
      piece.delta_center_to_top_left = null ;
      piece.delta_center_to_bottom_right = null ;
      piece.free_edges = null ;
      piece.connected_pieces = null ;
      piece.piece_image = null ;
      piece.free_edges = new Edge [4] ;
      /* Allocate memory for the 4 edges. */
      for (int k = 0; k < 4; k++) {
        piece.free_edges[k] = new Edge () ;
        if (piece.free_edges[k] == null) return (null) ;
        piece.free_edges[k].p0 = new Point () ;
        piece.free_edges[k].p1 = new Point () ;
      }    /* End of for (k = 0; k < 4; k++). */
    }

    /* Now, allocate the layout. Note that the layout must contain at least
       as many cells than the matrix of pieces has otherwise the pieces of
       the jigsaw would't fit inside.
       Moreover, to let the player move pieces on an area wider than the
       jigsaw's image, we make the layout bigger and use all the window's
       surface. */
    layout_nb_columns = screen_width / piece_width ;
    if ((screen_width % piece_width) != 0) layout_nb_columns++ ;
    layout_nb_columns = Math.max (layout_nb_columns, nb_columns_of_pieces) ;
    layout_nb_lines = screen_height / piece_height ;
    if ((screen_height % piece_height) != 0) layout_nb_lines++ ;
    layout_nb_lines = Math.max (layout_nb_lines, nb_lines_of_pieces) ;
    layout = new Layout_cell [layout_nb_columns * layout_nb_lines] ;

    /* Initialize the layout. */
    for (i = 0; i < (layout_nb_columns * layout_nb_lines); i++) {
      layout[i] = new Layout_cell () ;
      layout[i].hosted_pieces = null ;
    }

    /* Loop for basic creation of the pieces, some parts will be
       initialized to default value and update later when more
       information will be available. */
    for (x = 0; x < nb_columns_of_pieces; x++) {
      for (y = 0; y < nb_lines_of_pieces; y++) {
        int piece_index_and_id = (y * nb_columns_of_pieces) + x ;

        /* Keep a pointer to the current piece in the pieces matrix. */
        Piece piece = pieces[piece_index_and_id] ;
        /* Compute the id of the first edge of the piece. */
        int first_edge_debug_id = ((x * nb_lines_of_pieces) + y) * 4 ;
        /* Initialize the piece's structure. Note that id is equal to index
           in the array of pieces. */
        piece.piece_debug_id = piece_index_and_id ;
        piece.center = new Point () ;
        piece.center.x =  (x * piece_width) + (piece_width / 2) ;
        piece.center.y = (y * piece_height) + (piece_height / 2) ;
        /* Bounding box will be initialized once the splines are computed. For
           the moment, setting it to (0, 0 - (0, 0) is fine since this
           corresponds to a 0-wide bounding box centered around the piece's
           center. */
        piece.delta_center_to_top_left = new Point () ;
        piece.delta_center_to_top_left.x = 0 ;
        piece.delta_center_to_top_left.y = 0 ;
        piece.delta_center_to_bottom_right = new Point () ;
        piece.delta_center_to_bottom_right.x = 0 ;
        piece.delta_center_to_bottom_right.y = 0 ;
         
        /* To be fully initialized below once edges exist. */
        piece.free_edges[PieceStruct.NORTH].active = false ;
        piece.free_edges[PieceStruct.EAST].active = false ;
        piece.free_edges[PieceStruct.SOUTH].active = false ;
        piece.free_edges[PieceStruct.WEST].active = false ;
        /* Initialize the connected set with a fresh one only containing
           the current piece. */
        {
          Connected_set con_set ;
          Pieces_list p_lst ;

          con_set = new Connected_set () ;
          p_lst = new Pieces_list () ;

          p_lst.next = null ;
          p_lst.piece = piece ;
          con_set.members = p_lst ;
          con_set.refcount = 1 ;
          piece.connected_pieces = con_set ;
        }
        /* Image will be initialized later. */
        piece.piece_image = null ;

        /* Now, create the edges of the piece. */

        /* North. */
        piece.free_edges[PieceStruct.NORTH].active = true ;
        piece.free_edges[PieceStruct.NORTH].p0.x = x * piece_width ;
        piece.free_edges[PieceStruct.NORTH].p0.y = y * piece_height ;
        piece.free_edges[PieceStruct.NORTH].p0_anchor_id =
          x + y * (nb_columns_of_pieces + 1) ;
        piece.free_edges[PieceStruct.NORTH].p1.x = (x + 1) * piece_width - 1 ;
        piece.free_edges[PieceStruct.NORTH].p1.y = y * piece_height ;
        piece.free_edges[PieceStruct.NORTH].p1_anchor_id =
          1 + x + y * (nb_columns_of_pieces + 1) ;
        piece.free_edges[PieceStruct.NORTH].spline = null ; /* To be initialized later. */
        piece.free_edges[PieceStruct.NORTH].part_of_piece = piece ;
        piece.free_edges[PieceStruct.NORTH].x_sorted_prev = null ;
        piece.free_edges[PieceStruct.NORTH].x_sorted_next = null ;
        piece.free_edges[PieceStruct.NORTH].y_sorted_prev = null ;
        piece.free_edges[PieceStruct.NORTH].y_sorted_next = null ;
        piece.free_edges[PieceStruct.NORTH].id = first_edge_debug_id ;

        /* East. */
        piece.free_edges[PieceStruct.EAST].active = true ;
        piece.free_edges[PieceStruct.EAST].p0.x = (x + 1) * piece_width - 1 ;
        piece.free_edges[PieceStruct.EAST].p0.y = y * piece_height ;
        piece.free_edges[PieceStruct.EAST].p0_anchor_id =
          1 + x + y * (nb_columns_of_pieces + 1) ;
        piece.free_edges[PieceStruct.EAST].p1.x = (x + 1) * piece_width - 1 ;
        piece.free_edges[PieceStruct.EAST].p1.y = (y + 1) * piece_height - 1 ;
        piece.free_edges[PieceStruct.EAST].p1_anchor_id =
          1 + x + (y + 1) * (nb_columns_of_pieces + 1) ;
        piece.free_edges[PieceStruct.EAST].spline = null ; /* To be initialized later. */
        piece.free_edges[PieceStruct.EAST].part_of_piece = piece ;
        piece.free_edges[PieceStruct.EAST].x_sorted_prev = null ;
        piece.free_edges[PieceStruct.EAST].x_sorted_next = null ;
        piece.free_edges[PieceStruct.EAST].y_sorted_prev = null ;
        piece.free_edges[PieceStruct.EAST].y_sorted_next = null ;
        piece.free_edges[PieceStruct.EAST].id = first_edge_debug_id + 1 ;

        /* South. */
        piece.free_edges[PieceStruct.SOUTH].active = true ;
        piece.free_edges[PieceStruct.SOUTH].p0.x = (x + 1) * piece_width - 1 ;
        piece.free_edges[PieceStruct.SOUTH].p0.y = (y + 1) * piece_height - 1 ;
        piece.free_edges[PieceStruct.SOUTH].p0_anchor_id =
          1 + x + (y + 1) * (nb_columns_of_pieces + 1) ;
        piece.free_edges[PieceStruct.SOUTH].p1.x = x * piece_width ;
        piece.free_edges[PieceStruct.SOUTH].p1.y = (y + 1) * piece_height - 1 ;
        piece.free_edges[PieceStruct.SOUTH].p1_anchor_id =
          x + (y + 1) * (nb_columns_of_pieces + 1) ;
        piece.free_edges[PieceStruct.SOUTH].spline = null ; /* To be initialized later. */
        piece.free_edges[PieceStruct.SOUTH].part_of_piece = piece ;
        piece.free_edges[PieceStruct.SOUTH].x_sorted_prev = null ;
        piece.free_edges[PieceStruct.SOUTH].x_sorted_next = null ;
        piece.free_edges[PieceStruct.SOUTH].y_sorted_prev = null ;
        piece.free_edges[PieceStruct.SOUTH].y_sorted_next = null ;
        piece.free_edges[PieceStruct.SOUTH].id = first_edge_debug_id + 2 ;

        /* West. */
        piece.free_edges[PieceStruct.WEST].active = true ;
        piece.free_edges[PieceStruct.WEST].p0.x = x * piece_width ;
        piece.free_edges[PieceStruct.WEST].p0.y = (y + 1) * piece_height - 1 ;
        piece.free_edges[PieceStruct.WEST].p0_anchor_id =
          x + (y + 1) * (nb_columns_of_pieces + 1) ;
        piece.free_edges[PieceStruct.WEST].p1.x = x * piece_width ;
        piece.free_edges[PieceStruct.WEST].p1.y = y * piece_height ;
        piece.free_edges[PieceStruct.WEST].p1_anchor_id =
          x + y * (nb_columns_of_pieces + 1) ;
        piece.free_edges[PieceStruct.WEST].spline = null ; /* To be initialized later. */
        piece.free_edges[PieceStruct.WEST].part_of_piece = piece ;
        piece.free_edges[PieceStruct.WEST].x_sorted_prev = null ;
        piece.free_edges[PieceStruct.WEST].x_sorted_next = null ;
        piece.free_edges[PieceStruct.WEST].y_sorted_prev = null ;
        piece.free_edges[PieceStruct.WEST].y_sorted_next = null ;
        piece.free_edges[PieceStruct.WEST].id = first_edge_debug_id + 3 ;
      }    /* End of for (y = 0; y < nb_lines; y++). */
    }      /* End of for (x = 0; x < nb_columns; x++). */

    /* Now, splines updates for all pieces. */

    /* Update the splines north and south by splitting by columns. */
    for (x = 0; x < nb_columns_of_pieces; x++) {
      /* Just used to store the index of a piece in the pieces matrix. */
      int index ;
      /* Some temporary variables. */
      Point[] tmp_point_ptr ;
      Spline tmp_spline_ptr ;

      /* Init of the north of each cell of the first line. */
      index = x ;       /* pieces[x][0]. */
      tmp_point_ptr = new Point [2] ;
      tmp_point_ptr[0] = new Point () ;
      tmp_point_ptr[1] = new Point () ;
      tmp_spline_ptr = new Spline () ;
      tmp_spline_ptr.nb_points = 2 ;
      tmp_spline_ptr.vertices = tmp_point_ptr ;
      pieces[index].free_edges[PieceStruct.NORTH].spline = tmp_spline_ptr ;
      tmp_point_ptr[0].x = x * piece_width ;
      tmp_point_ptr[0].y = 0 ;
      tmp_point_ptr[1].x = (x + 1) * piece_width - 1 ;
      tmp_point_ptr[1].y = 0 ;
      pieces[index].delta_center_to_top_left.x = - (piece_width / 2) ;
      pieces[index].delta_center_to_top_left.y = - (piece_height / 2) ;
      pieces[index].delta_center_to_bottom_right.x = piece_width / 2 ;
      pieces[index].delta_center_to_bottom_right.y = - (piece_height / 2) ;

      /* Init of the south of each cell of the last line. */
      /* pieces[x][nb_lines - 1]. */
      index = ((nb_lines_of_pieces - 1) * nb_columns_of_pieces) + x ;
      tmp_point_ptr = new Point [2] ;
      tmp_point_ptr[0] = new Point () ;
      tmp_point_ptr[1] = new Point () ;
      tmp_spline_ptr = new Spline () ;
      tmp_spline_ptr.nb_points = 2 ;
      tmp_spline_ptr.vertices = tmp_point_ptr ;
      pieces[index].free_edges[PieceStruct.SOUTH].spline = tmp_spline_ptr ;
      tmp_point_ptr[0].x = (x + 1) * piece_width - 1 ;
      tmp_point_ptr[0].y = image_height - 1 ;
      tmp_point_ptr[1].x = x * piece_width ;
      tmp_point_ptr[1].y = image_height - 1 ;
      pieces[index].delta_center_to_top_left.x = - (piece_width / 2) ;
      pieces[index].delta_center_to_top_left.y = piece_height / 2 ;
      pieces[index].delta_center_to_bottom_right.x = piece_width / 2 ;
      pieces[index].delta_center_to_bottom_right.y = piece_height / 2 ;

      /* Create the horizontal splines of the column. */
      for (y = 1; y < nb_lines_of_pieces; y++) {
        /* Points used as out-mode parameter to receive the spline's
          bounding-box when a spline is created. */
        Point spline_top_left = new Point () ;
        Point spline_bottom_right = new Point () ;
        Point [] horiz_segments ;
        Spline horiz_spline ;

        /* Horizontal spline, hence north of the current cell. */
        horiz_segments =
          PieceStruct.makeHorizontalEdgeSegments
          ((x * piece_width), (y * piece_height), (piece_width - 1)) ;
        if (horiz_segments == null) return (null) ;

        horiz_spline =
          SplineStruct.buildSpline
          (1, horiz_segments, PieceStruct.NB_UNIT_EDGE_POINTS, spline_top_left,
              spline_bottom_right) ;
        if (horiz_spline == null) return (null) ;

        /* Release the segments, they are useless now. */
        horiz_segments = null ;

        /* Init the north of the current cell. */
        index = (y * nb_columns_of_pieces) + x ;    /* pieces [x][y]. */
        pieces[index].free_edges[PieceStruct.NORTH].spline = horiz_spline ;
        pieces[index].delta_center_to_top_left.x =
          Math.min
          (pieces[index].delta_center_to_top_left.x,
              (spline_top_left.x - pieces[index].center.x)) ;
        pieces[index].delta_center_to_top_left.y =
          Math.min
          (pieces[index].delta_center_to_top_left.y,
              (spline_top_left.y - pieces[index].center.y)) ;
        pieces[index].delta_center_to_bottom_right.x =
          Math.max
          (pieces[index].delta_center_to_bottom_right.x,
              (spline_bottom_right.x - pieces[index].center.x)) ;
        pieces[index].delta_center_to_bottom_right.y =
          Math.max
          (pieces[index].delta_center_to_bottom_right.y,
              (spline_bottom_right.y - pieces[index].center.y)) ;

        /* Init of the south of the y previous cell. */
        if (y > 0) {
          /* pieces[x][y - 1]. */
          index = ((y - 1) * nb_columns_of_pieces) + x ;
          pieces[index].free_edges[PieceStruct.SOUTH].spline =
            SplineStruct.build_opposite_spline (/* shift_up: */ true, horiz_spline) ;
          pieces[index].delta_center_to_top_left.x =
            Math.min
            (pieces[index].delta_center_to_top_left.x,
                (spline_top_left.x - pieces[index].center.x)) ;
          pieces[index].delta_center_to_top_left.y =
            Math.min
            (pieces[index].delta_center_to_top_left.y,
                (spline_top_left.y - pieces[index].center.y)) ;
          pieces[index].delta_center_to_bottom_right.x =
            Math.max
            (pieces[index].delta_center_to_bottom_right.x,
                (spline_bottom_right.x - pieces[index].center.x)) ;
          pieces[index].delta_center_to_bottom_right.y =
            Math.max
            (pieces[index].delta_center_to_bottom_right.y,
                (spline_bottom_right.y - pieces[index].center.y)) ;
        }   /* End of if (y > 0). */
      }     /* End of for (y = 1; y < nb_lines; y++). */
    }       /* End of for (x = 0; x < to nb_columns; x++). */

    /* Update the splines east and west by splitting by lines. */
    for (y = 0; y < nb_lines_of_pieces; y++) {
      /* Just used to store the index of a piece in the pieces matrix. */
      int index ;
      /* Some temporary variables. */
      Point[] tmp_point_ptr ;
      Spline tmp_spline_ptr ;

      /* Init of the west of each cell of the first column. */
      /* pieces[0][y]. */
      index = y * nb_columns_of_pieces ;
      tmp_point_ptr = new Point [2] ;
      tmp_point_ptr[0] = new Point () ;
      tmp_point_ptr[1] = new Point () ;
      tmp_spline_ptr = new Spline () ;
      tmp_spline_ptr.nb_points = 2 ;
      tmp_spline_ptr.vertices = tmp_point_ptr ;
      pieces[index].free_edges[PieceStruct.WEST].spline = tmp_spline_ptr ;
      tmp_point_ptr[0].x = 0 ;
      tmp_point_ptr[0].y = (y + 1) * piece_height - 1 ;
      tmp_point_ptr[1].x = 0 ;
      tmp_point_ptr[1].y = y * piece_height ;
      pieces[index].delta_center_to_top_left.x =
        Math.min
        (pieces[index].delta_center_to_top_left.x,
            (- (piece_width / 2))) ;
      pieces[index].delta_center_to_top_left.y =
        Math.min
        (pieces[index].delta_center_to_top_left.y,
            (- (piece_height / 2))) ;
      pieces[index].delta_center_to_bottom_right.x =
        Math.max
        (pieces[index].delta_center_to_bottom_right.x,
            (- (piece_width / 2))) ;
      pieces[index].delta_center_to_bottom_right.y =
        Math.max
        (pieces[index].delta_center_to_bottom_right.y,
            (piece_height / 2)) ;
     
      /* Init of the east of each cell of the last column. */
      /* pieces[nb_columns - 1][y]. */
      index = (y * nb_columns_of_pieces) + (nb_columns_of_pieces - 1) ;
      tmp_point_ptr = new Point [2] ;
      tmp_point_ptr[0] = new Point () ;
      tmp_point_ptr[1] = new Point () ;
      tmp_spline_ptr = new Spline () ;
      tmp_spline_ptr.nb_points = 2 ;
      tmp_spline_ptr.vertices = tmp_point_ptr ;
      pieces[index].free_edges[PieceStruct.EAST].spline = tmp_spline_ptr ;
      tmp_point_ptr[0].x = image_width - 1 ;
      tmp_point_ptr[0].y = y * piece_height ;
      tmp_point_ptr[1].x = image_width - 1 ;
      tmp_point_ptr[1].y = (y + 1) * piece_height - 1 ;
      pieces[index].delta_center_to_top_left.x =
        Math.min
        (pieces[index].delta_center_to_top_left.x,
            (piece_width / 2)) ;
      pieces[index].delta_center_to_top_left.y =
        Math.min
        (pieces[index].delta_center_to_top_left.y,
            (- (piece_height / 2))) ;
      pieces[index].delta_center_to_bottom_right.x =
        Math.max
        (pieces[index].delta_center_to_bottom_right.x,
            (piece_width / 2)) ;
      pieces[index].delta_center_to_bottom_right.y =
        Math.max
        (pieces[index].delta_center_to_bottom_right.y,
            (piece_height / 2)) ;

      /* Create the vertical splines of the line. */
      for (x = 1; x < nb_columns_of_pieces; x++) {
        /* Points used as out-mode parameter to receive the spline's
          bounding-box when a spline is created. */
        Point spline_top_left = new Point () ;
        Point spline_bottom_right = new Point () ;
        Point[] vert_segments ;
        Spline vert_spline ;

        /* Vertical spline, hence west of the current cell. */
        vert_segments =
          PieceStruct.makeVerticalEdgeSegments
          ((x * piece_width), (y * piece_height), (piece_height - 1)) ;
        if (vert_segments == null) return (null) ;
       
        vert_spline =
          SplineStruct.buildSpline
          (1, vert_segments, PieceStruct.NB_UNIT_EDGE_POINTS, spline_top_left,
              spline_bottom_right) ;
        if (vert_spline == null) return (null) ;

        /* Release the segments, they are useless now. */
        vert_segments = null ;

        /* Initialise the west of the current cell. */
        /* pieces [x][y]. */
        index = (y * nb_columns_of_pieces) + x ;
        pieces[index].free_edges[PieceStruct.WEST].spline = vert_spline ;
        pieces[index].delta_center_to_top_left.x =
          Math.min
          (pieces[index].delta_center_to_top_left.x,
              (spline_top_left.x - pieces[index].center.x)) ;
        pieces[index].delta_center_to_top_left.y =
          Math.min
          (pieces[index].delta_center_to_top_left.y,
              (spline_top_left.y - pieces[index].center.y)) ;
        pieces[index].delta_center_to_bottom_right.x =
          Math.max
          (pieces[index].delta_center_to_bottom_right.x,
              (spline_bottom_right.x - pieces[index].center.x)) ;
        pieces[index].delta_center_to_bottom_right.y =
          Math.max
          (pieces[index].delta_center_to_bottom_right.y,
              (spline_bottom_right.y - pieces[index].center.y)) ;

        /* Init of the east of the previous x cell. */
        if (x > 0) {
          /* piece[x-1][y]. */
          index = (y * nb_columns_of_pieces) + (x - 1) ;
          pieces[index].free_edges[PieceStruct.EAST].spline =
            SplineStruct.build_opposite_spline
            	(/* shift_up: */ false, vert_spline) ;
          pieces[index].delta_center_to_top_left.x =
            Math.min
            (pieces[index].delta_center_to_top_left.x,
                (spline_top_left.x - pieces[index].center.x)) ;
          pieces[index].delta_center_to_top_left.y =
            Math.min
            (pieces[index].delta_center_to_top_left.y,
                (spline_top_left.y - pieces[index].center.y)) ;

          pieces[index].delta_center_to_bottom_right.x =
            Math.max
            (pieces[index].delta_center_to_bottom_right.x,
                (spline_bottom_right.x - pieces[index].center.x)) ;
          pieces[index].delta_center_to_bottom_right.y =
            Math.max
            (pieces[index].delta_center_to_bottom_right.y,
                (spline_bottom_right.y - pieces[index].center.y)) ;
        }    /* End of if (x > 0). */
      }      /* End of for (x = 1; x < nb_columns; x++). */
    }        /* End of for (y = 0; y < nb_lines; y++). */

    /* Initialize the sorted double linked lists of increasing x. */
    prev = null ;
    for (x = 0; x < nb_columns_of_pieces; x++) {
      for (y = 0; y < nb_lines_of_pieces; y++) {
        /* Processing north edge of the current cell (pieces[x][y]). */
        tmp_edge =
        	pieces[(y * nb_columns_of_pieces) + x].free_edges[PieceStruct.NORTH] ;
        if (prev != null) insertAfterInXList (tmp_edge, prev) ;
        prev = tmp_edge ;
        /* Processing east edge of the left cell if some (pieces[x-1][y]). */
        if (x > 0) {
          tmp_edge =
            pieces[(y * nb_columns_of_pieces) + (x - 1)].free_edges[PieceStruct.EAST] ;
          if (prev != null) insertAfterInXList (tmp_edge, prev) ;
          prev = tmp_edge ;
        }
        /* Processing the west edge of the current cell (pieces[x][y]). */
        tmp_edge =
        	pieces[(y * nb_columns_of_pieces) + x].free_edges[PieceStruct.WEST] ;
        if (prev != null) insertAfterInXList (tmp_edge, prev) ;
        prev = tmp_edge ;
        /* Processing the south edge of the current cell (pieces[x][y]). */
        tmp_edge =
        	pieces[(y * nb_columns_of_pieces) + x].free_edges[PieceStruct.SOUTH] ;
        if (prev != null) insertAfterInXList (tmp_edge, prev) ;
        prev = tmp_edge ;
      }   /* End of for (y = 0; y < nb_lines; y++). */
    }     /* End of for (x = 0; x < nb_columns; x++). */

    /* Processing the east edges of the last column (pieces[nb_columns-1][y]). */
    for (y = 0; y < nb_lines_of_pieces; y++) {
      tmp_edge =
        pieces[(y * nb_columns_of_pieces) + (nb_columns_of_pieces - 1)].
        free_edges[PieceStruct.EAST] ;
      if (prev != null) insertAfterInXList (tmp_edge, prev) ;
      prev =tmp_edge ;
    }

    /* Initialize the sorted double linked lists of increasing y. */
    prev = null ;
    for (y = 0; y < nb_lines_of_pieces; y++) {
      for (x = 0 ; x < nb_columns_of_pieces; x++) {
        /* Processing south edge of the above cell if some (pieces[x][y-1]). */
        if (y > 0) {
          tmp_edge =
            pieces[((y - 1) * nb_columns_of_pieces) + x].free_edges[PieceStruct.SOUTH] ;
          if (prev != null) InsertAfterInYList (tmp_edge, prev) ;
          prev = tmp_edge ;
        }
        /* Processing north edge of the current cell (pieces[x][y]). */
        tmp_edge =
        	pieces[(y * nb_columns_of_pieces) + x].free_edges[PieceStruct.NORTH] ;
        if (prev != null) InsertAfterInYList (tmp_edge, prev) ;
        prev = tmp_edge ;
        /* Processing the west edge of the current cell (pieces[x][y]). */
        tmp_edge =
        	pieces[(y * nb_columns_of_pieces) + x].free_edges[PieceStruct.WEST] ;
        if (prev != null) InsertAfterInYList (tmp_edge, prev) ;
        prev = tmp_edge ;
        /* Processing east edge of the current cell (pieces[x][y]). */
        tmp_edge =
        	pieces[(y * nb_columns_of_pieces) + x].free_edges[PieceStruct.EAST] ;
        if (prev != null) InsertAfterInYList (tmp_edge, prev) ;
        prev = tmp_edge ;
      }   /* End of for (x = 0 ; x < nb_columns; x++). */
    }     /* End of for (y = 0; y <nb_lines; y++). */

    /* Processing the south edges of the last line (pieces[x][nb_lines - 1]). */
    for (x = 0; x < nb_columns_of_pieces; x++) {
      tmp_edge =
        pieces[((nb_lines_of_pieces - 1) * nb_columns_of_pieces) + x].
        free_edges[PieceStruct.SOUTH] ;
      if (prev != null) InsertAfterInYList (tmp_edge, prev) ;
      prev = tmp_edge ;
    }

    /* Create the game structure. */
    game = new Game () ;

    game.rotated_image = rotate ;
    game.game_won = false ;
    game.image_width = image_width ;
    game.image_height = image_height ;
    game.original_image_scale_factor = scale_factor ;
    game.piece_width = piece_width ;
    game.piece_height = piece_height ;
    game.layout_nb_lines = layout_nb_lines ;
    game.layout_nb_columns = layout_nb_columns ;
    game.layout = layout ;
    game.pieces = pieces ;
    game.nb_pieces = nb_columns_of_pieces * nb_lines_of_pieces ;
    /* By construction, the first edge is north of the first piece ([0][0]). */
    game.x_sorted_edges_head = pieces[0].free_edges[PieceStruct.NORTH] ;
    /* By construction, the last edge is east of the last piece
      ([nb_columns - 1][nb_lines - 1]) . */
    game.x_sorted_edges_tail =
      pieces
      [((nb_lines_of_pieces - 1) * nb_columns_of_pieces) +
       (nb_columns_of_pieces - 1)].free_edges[PieceStruct.EAST] ;
    /* By construction, the first edge is north of the first piece ([0][0]). */
    game.y_sorted_edges_head = pieces[0].free_edges[PieceStruct.NORTH] ;
    /* By construction, the last edge is south of the last piece
       ([nb_columns - 1][nb_lines - 1]). */
    game.y_sorted_edges_tail =
      pieces
      [((nb_lines_of_pieces - 1) * nb_columns_of_pieces) +
       (nb_columns_of_pieces - 1)].free_edges[PieceStruct.SOUTH] ;
    game.game_image = game_image ;
    game.image_file_size = raw_image_file_size ;
    game.image_file_data = raw_image_file_data ;

    /* Now bounding-boxes are computed, and we have a game structure
       we can record the pieces in the layout and by the way texture
       map them. */
    for (x = 0; x < nb_columns_of_pieces; x++) {
      for (y = 0; y < nb_lines_of_pieces; y++) {
        /* pieces[x][y]. */
        Piece piece = pieces[(y * nb_columns_of_pieces) + x] ;

        if (! GameManips.insertPieceInLayout (game, piece)) return (null) ;

        /* Since the game is not yet shuffled, pieces are still positionned
          exactly matching their position in the game picture. So we can directly
          use the center position that is recorded in the piece itself. */
        if (! TextureMap.initializePieceTexture (game.game_image, piece, piece.center)) {
          return (null) ;
        }
      }    /* End of for (y = 0; y < nb_lines; y++). */
    }      /* End of for (x = 0; x < nb_columns; x++). */

    /* Finally, return the game structure. */
    return (game) ;
  }

}
