package francois.pessaux.androjig.android;

import java.io.DataInputStream ;
import java.io.DataOutputStream ;
import java.io.File ;
import java.io.FileInputStream ;
import java.io.FileOutputStream ;
import java.io.IOException ;

import android.graphics.Bitmap ;
import android.os.Environment ;

import francois.pessaux.androjig.android.GameStruct.Game ;
import francois.pessaux.androjig.android.GameStruct.Layout_cell ;
import francois.pessaux.androjig.android.Maths2d.Point ;
import francois.pessaux.androjig.android.PieceStruct.Connected_set ;
import francois.pessaux.androjig.android.PieceStruct.Edge ;
import francois.pessaux.androjig.android.PieceStruct.Piece ;
import francois.pessaux.androjig.android.PieceStruct.Pieces_list ;
import francois.pessaux.androjig.android.SplineStruct.Spline ;


public class GameLoad {
  /** Local temporary structure used to rebuild connected sets circumvening the
      fact that when loading them from the saved game, pieces are not yet
      loaded. */
  private static class Pre_connected_set {
    Connected_set partial_connected_set ;
    /* Number of pieces members of the pre-connected-set, hence of the related
       future connected-set. */
    int nb_members ;
    /* An array containing the ids of the pieces members of this
       pre-connected-set, hence of the related future connected-set. */
    int[] members_ids ;
  };



  private static class Pre_edge {
    Edge partial_edge ;
    int future_edge_part_of_piece_id ; // TODO It seems we never read it.
    int future_edge_x_sorted_prev_id ;
    int future_edge_x_sorted_next_id ;
    int future_edge_y_sorted_prev_id ;
    int future_edge_y_sorted_next_id ;
  };



  /** Remember that the array of pieces is expected to be indexed by the pieces
      ids. 
   * @throws IOException */
  private static Layout_cell[] loadLayout (DataInputStream in, int nb_cells,
                                           Piece[] pieces)
    throws IOException
  {
    int i ;
    Layout_cell[] layout ;

    /* Allocate the layout itself. */
    layout = new Layout_cell [nb_cells] ;
    for (i = 0; i < nb_cells; i++) {
      layout[i] = new Layout_cell () ;
      layout[i].hosted_pieces = null ;
    }

    
    /* Load the list of pieces ids of each layout cell. Set the redraw flag
       to false because we will anyway redraw the whole screen at startup. */
    for (i = 0; i < nb_cells; i++) {
      Pieces_list list_head = null ;
      int nb_pieces, j ;
      int ids[] ;

      /* Read the number of pieces in the list. */
      nb_pieces = in.readInt () ;

      /* Now, read all the ids in one shot. */
      ids = new int [nb_pieces] ;

      for (j = 0; j < nb_pieces; j++) ids[j] = in.readInt () ;
  
      /* Rebuild the list starting by the end of the array to be able to add in
         head, which is more efficient. */
      for (j = nb_pieces - 1; j >= 0; j--) {
        Pieces_list new_cell = new Pieces_list () ;
        new_cell.piece = pieces[ids[j]] ;
        /* Insert in head of the list. */
        new_cell.next = list_head ;
        list_head = new_cell ;
      }    /* End of for (j = nb_cells - 1; j >= 0; j--). */

      layout[i].hosted_pieces = list_head ;
    }     /* End of for (i = 0; i < nb_cells; i++). */

    return (layout) ;
  }


  
  private static Point loadPoint (DataInputStream in)
    throws IOException
  {
    Point point = new Point () ;
    point.x = in.readInt () ;
    point.y = in.readInt () ;
    return (point) ;
  }
  
  
  
  private static Spline loadSpline (DataInputStream in)
    throws IOException
  {
    Spline spline = new Spline () ;

    /* First, read the number of points. */
    spline.nb_points = in.readInt () ;

    spline.vertices = new Point [spline.nb_points] ;
    if (spline.vertices == null) return (null) ;

    for (int i = 0; i < spline.nb_points; i++)
      spline.vertices[i] = loadPoint (in) ;

    return (spline) ;
  }



  /** Create an array containing the pre-edges (that have a pointer on the
      allocated edge) from the saved edges description. These pre-edges contain
      partially initialized edges, in particular their links in the
      double-linked X/Y sorted lists are not set and the piece the edge is part
      of is not set. These pre-edges also contain material (ids) to be able later
      to fully initialize the edges pointers.
      The aim of this function is to pre-create the edges, store them in an
      array indexed by edges ids in order to be able to have all the addresses
      of edges availables to later create pieces and link pieces in the double
      linked X/Y sorted lists and in the pieces.
      In fact, since we have mutualm references between edges and between edges
      and pieces, we can't rebuild the data-structures in one pass. 
   * @throws IOException */
  private static Pre_edge[] loadAllEdgesAndMakePreEdges (DataInputStream in,
                                                         int nb_edges)
    throws IOException
  {
    int i ;
    Pre_edge[] all_pre_edges = null ;

    /* Allocate the array to remind the edges pointers. Once pieces and edges are
       fully initialized, this array can be freed (but not the edges the
       pre-edges have pointers on, of course, since they are retained in the
       pieces and other edges data-structures. */
    all_pre_edges = new Pre_edge [nb_edges] ;

    for (i = 0; i < nb_edges; i++) {
      all_pre_edges[i] = new Pre_edge () ;
      all_pre_edges[i].future_edge_part_of_piece_id = 0 ;
      all_pre_edges[i].future_edge_x_sorted_prev_id = 0 ;
      all_pre_edges[i].future_edge_x_sorted_next_id = 0 ;
      all_pre_edges[i].future_edge_y_sorted_prev_id = 0 ;
      all_pre_edges[i].future_edge_y_sorted_next_id = 0 ;
      all_pre_edges[i].partial_edge = null ;
    }

    for (i = 0; i < nb_edges; i++) {
      Edge edge ;
      Pre_edge curr_pre_edge ;

      edge = new Edge () ;

      /* Read the edge id. */
      edge.id = in.readInt () ;

      /* Store the edge in the array pre-edge cell at the index corresponding to
         its id and eemind the address of the pre-edge we are building. */
      all_pre_edges[edge.id].partial_edge = edge ;
      curr_pre_edge = all_pre_edges[edge.id] ;

      /* Edge activity. */
      edge.active = in.readBoolean () ;
      /* Point p0. */
      edge.p0 = loadPoint (in) ;
      /* Point p0 anchor id. */
      edge.p0_anchor_id = in.readInt () ;
      /* Point p1. */
      edge.p1 = loadPoint (in) ;
      /* Point p1 anchor id. */
      edge.p1_anchor_id = in.readInt () ;

      /* The spline. */
      edge.spline = loadSpline (in) ;
      if (edge.spline == null) return (null) ;

      /* The id of the piece this edge belongs to. */
      curr_pre_edge.future_edge_part_of_piece_id = in.readInt () ;
      /* The ids of our next/prev edges in the X/Y sorted lists. */
      curr_pre_edge.future_edge_x_sorted_prev_id = in.readInt () ;
      curr_pre_edge.future_edge_x_sorted_next_id = in.readInt () ;
      curr_pre_edge.future_edge_y_sorted_prev_id = in.readInt () ;
      curr_pre_edge.future_edge_y_sorted_next_id = in.readInt () ;
    }    /* End of for (i = 0; i < nb_edges; i++). */

    return (all_pre_edges) ;
  }

  
  
  /** Remember that the array of pieces is expected to be indexed by the pieces
      ids. */
  private static void finalizeConnectedSets
      (Pre_connected_set[] pre_connected_sets,
       int nb_connected_sets, Piece[] pieces, int nb_pieces)
  {
    int i ;

    /* Now the pieces are well-ordered in the array, we can finalize the
       connected-sets one by one. */
    for (i = 0; i < nb_connected_sets; i++) {
      int j ;
      Pieces_list list_head = null ;

      /* Rebuild the list starting by the end of the array to be able to add in
         head, which is more efficient. */
      for (j = pre_connected_sets[i].nb_members - 1; j >= 0; j--) {
        Pieces_list new_cell = new Pieces_list () ;
        new_cell.piece = pieces[pre_connected_sets[i].members_ids[j]] ;
        /* Insert in head of the list. */
        new_cell.next = list_head ;
        list_head = new_cell ;
      }    /* End of for (j = pre_connected_sets[i].nb_members - 1; ...). */

      /* Finally, record the head of the list. */
      pre_connected_sets[i].partial_connected_set.members = list_head ;
    }   /* End of for (i = 0; i < nb_connected_sets; i++). */
  }

  
  /** Terminates the edges initialization by updating the double-linked X/Y
      sorted lists. Until all the edges were created, we didn't have all the
      pointers on the edges. When this function is called, we assume we have all
      these pointers. Hence, we can finally make the pointers pointing at the
      right places in these lists in the edges.
      The array of pre-edges we receive is expected to be indexed by the edges
      ids. Hence when searching an edge with a certain id, we will find it in
      the array at the index equal to this id. */
  private static void finalizeEdgesLinkedLists (Pre_edge[] pre_edges, int nb_edges)
  {
    int i ;

    for (i = 0; i < nb_edges; i++) {
      Pre_edge pre_edge = pre_edges[i] ;
      Edge edge = pre_edge.partial_edge ;

      /* Deal with the x_sorted_prev. */
      edge.x_sorted_prev =
        ((pre_edge.future_edge_x_sorted_prev_id == -1) ?
            null : pre_edges[pre_edge.future_edge_x_sorted_prev_id].partial_edge) ;
      /* Deal with the x_sorted_next. */
      edge.x_sorted_next =
        ((pre_edge.future_edge_x_sorted_next_id == -1) ?
            null : pre_edges[pre_edge.future_edge_x_sorted_next_id].partial_edge) ;
      /* Deal with the y_sorted_prev. */
      edge.y_sorted_prev =
        ((pre_edge.future_edge_y_sorted_prev_id == -1) ?
            null : pre_edges[pre_edge.future_edge_y_sorted_prev_id].partial_edge) ;
      /* Deal with the y_sorted_next. */
      edge.y_sorted_next =
        ((pre_edge.future_edge_y_sorted_next_id == -1) ?
            null : pre_edges[pre_edge.future_edge_y_sorted_next_id].partial_edge) ;
    }    /* End of for (i = 0; i < nb_edges; i++). */
  }

  
  
  private static class LoadPiecesResult {
    Piece[] pieces ;
    Edge x_head ;
    Edge x_tail ;
    Edge y_head ;
    Edge y_tail ;
  };
  
  
  
  /** x_head, x_tail, y_head and y_tail are out-mode parameters. 
   * @throws IOException */
  private static LoadPiecesResult loadPieces (DataInputStream in, int nb_pieces,
                                              int x_head_id, int x_tail_id,
                                              int y_head_id, int y_tail_id)
    throws IOException
  {
  int nb_connected_sets, i ;
  Pre_edge[] all_pre_edges = null ;
  Piece[] pieces = null ;
  Pre_connected_set[] pre_connected_sets = null ;

  /* We first read the number of saved connected-sets. */
  nb_connected_sets = in.readInt () ;

  pre_connected_sets = new Pre_connected_set [nb_connected_sets] ;
  
  for (i = 0; i < nb_connected_sets; i++) {
    pre_connected_sets[i] = new Pre_connected_set () ;
    pre_connected_sets[i].members_ids = null ;
    pre_connected_sets[i].nb_members = 0 ;
    pre_connected_sets[i].partial_connected_set = null ;
  }

  /* Now read the saved connected-sets to make pre-connected-sets. */
  for (i = 0; i < nb_connected_sets; i++) {
    /* The number of pieces members of this connected-set. */
    pre_connected_sets[i].nb_members = in.readInt () ;
    pre_connected_sets[i].members_ids =
      new int [pre_connected_sets[i].nb_members] ;
    if (pre_connected_sets[i].members_ids == null) {
      return (null) ;
    }
    
    /* Create the related real connected-set structure. */
    pre_connected_sets[i].partial_connected_set = new Connected_set () ;
    if (pre_connected_sets[i].partial_connected_set == null) {
      return (null) ;
    }
    
    /* Load the refcount directly in the connected-set structure. */
    pre_connected_sets[i].partial_connected_set.refcount = in.readInt () ;
    
    /* Load the ids. */
    for (int j = 0; j < pre_connected_sets[i].nb_members; j++)
      pre_connected_sets[i].members_ids[j] = in.readInt () ;
  }

  /* Ok, right, connected-sets information is loaded, future connected-sets
     structures are allocated, pre-connected-sets are built, we will have later
     to complete connected-sets structures by making lists of pieces whose
     indices are in the pre-connected-sets structures. But we will be able to
     do thing only once pieces are loaded and created.
     Now, address loading edges. */
  all_pre_edges = loadAllEdgesAndMakePreEdges (in, (nb_pieces * 4)) ;
  if (all_pre_edges == null) return (null) ;

  /* Now, address loading pieces information. */
  pieces = new Piece [nb_pieces] ;
  
  for (i = 0; i < nb_pieces; i++) {
    int tmp, j ;
    Piece piece = new Piece () ;
    pieces[i] = piece ;
    
    /* Read the id. */
    piece.piece_debug_id = in.readInt () ;
    /* The center. */
    piece.center = loadPoint (in) ;
    /* The bounding-box. */
    piece.delta_center_to_top_left = loadPoint (in) ;
    piece.delta_center_to_bottom_right = loadPoint (in) ;

    /* The edges. We assign the edges in the piece and by the way, we update the
       field part_of_piece of the related edge. */
    piece.free_edges = new Edge [4] ;
    for (j = 0; j < 4; j++) {
      tmp = in.readInt () ;
      piece.free_edges[j] = all_pre_edges[tmp].partial_edge ;
      piece.free_edges[j].part_of_piece = piece ;
    }

    /* Recover the connected-set this piece belongs to. We read the index of
       the connected-set. */
    tmp = in.readInt () ;
    piece.connected_pieces = pre_connected_sets[tmp].partial_connected_set ;
    
    /* The texture will be rebuilt later, once the picture is loaded. */
    piece.piece_image = null ;
  }     /* End of for (i = 0; i < nb_pieces; i++). */

  /* Now, terminate the edges initialization by setting the right pointers in
     the double-linked X/Y sorted lists inside the pieces. Since the field
     part_of_piece of each edge has been initialized while creating the pieces,
     after having initialized the linked lists, edges should be finally fully
     initialized. */
  finalizeEdgesLinkedLists (all_pre_edges, (nb_pieces * 4)) ;

  LoadPiecesResult result = new LoadPiecesResult () ;
  /* Set what are the pointer on heads and tails of the double-linked X/Y sorted
     lists. */
  result.x_head =
    ((x_head_id == -1) ? null : all_pre_edges[x_head_id].partial_edge) ;
  result.x_tail =
    ((x_tail_id == -1) ? null : all_pre_edges[x_tail_id].partial_edge) ;
  result.y_head =
    ((y_head_id == -1) ? null : all_pre_edges[y_head_id].partial_edge) ;
  result.y_tail =
    ((y_tail_id == -1) ? null : all_pre_edges[y_tail_id].partial_edge) ;

  result.pieces = pieces ;
  
  /* We don't need anymore the array of pre-edges, so free it. */
  all_pre_edges = null ;

  /* Now the pieces are created set the pointers to pieces in the
     connected-sets. */
  finalizeConnectedSets
    (pre_connected_sets, nb_connected_sets, pieces, nb_pieces) ;

  /* We dont need anymore the pre-connected-sets and the array storing the
     pointer to the connected-sets, so free them. */
  pre_connected_sets = null ;

  return (result) ;
  }

  
  
  private static Bitmap rebuildImageFromSavedBytes (byte[] data,
                                                    int size,
                                                    float scale_factor,
                                                    boolean rotated_image)
    throws IOException
  {
    File tmp_file = File.createTempFile ("androjig",".tmp") ;
    if (tmp_file == null) return (null) ;
    DataOutputStream tmp_out =
      new DataOutputStream (new FileOutputStream (tmp_file)) ;
    tmp_out.write (data, 0, size) ;
    tmp_out.close () ;

    Bitmap image =
      GameAlloc.loadGameImage
      	(tmp_file.getPath (), scale_factor, null, rotated_image) ;
    tmp_file.delete () ;

    return (image) ;  
  }

  
  
  public static Game loadGame (String file_name)
    throws IOException
  {
    int tmp ;
    int x_head_id, x_tail_id, y_head_id, y_tail_id ;
    int nb_columns_of_pieces, nb_lines_of_pieces, x, y ;

    File sdCard = Environment.getExternalStorageDirectory () ;
    File in_file = new File (sdCard, file_name) ;
    DataInputStream in = new DataInputStream (new FileInputStream (in_file)) ;

    /* First, check the magic number. */
    tmp = in.readInt () ;
    if (tmp != GameStruct.SAVEMAGICNUMBER) {
      in.close () ;
      return (null) ;
    }

  /* Allocate memory for the main game structure and clear all fields. */
  Game game = new Game () ;

  game.rotated_image = in.readBoolean () ;
  game.game_won = false ;

  /* Read the miscellaneous scalars. */
  game.image_width = in.readInt () ;
  game.image_height = in.readInt () ;
  game.original_image_scale_factor = in.readFloat () ;
  game.piece_width = in.readInt () ;
  game.piece_height = in.readInt () ;
  game.layout_nb_columns = in.readInt () ;
  game.layout_nb_lines = in.readInt () ;
  game.nb_pieces = in.readInt () ;

  /* The ids of the heads of the x and y sorted lists of edges or -1 if NULL. */
  x_head_id = in.readInt () ;
  x_tail_id = in.readInt () ;
  y_head_id = in.readInt () ;
  y_tail_id = in.readInt () ;

  /* Now, it's time to load the pieces. By the way, recover by side effect the
     addresses of the edges heads and tails of the double-linked X/Y sorted
     lists. */
  LoadPiecesResult lpr =
    loadPieces
      (in, game.nb_pieces, x_head_id, x_tail_id, y_head_id, y_tail_id) ;
  if (lpr == null) {
    in.close () ;
    return (null) ;
  }
  game.pieces = lpr.pieces ;
  
  game.x_sorted_edges_head = lpr.x_head ;
  game.x_sorted_edges_tail = lpr.x_tail ;
  game.y_sorted_edges_head = lpr.y_head ;
  game.y_sorted_edges_tail = lpr.y_tail ;

  /* Load the layout. */
  game.layout =
    loadLayout
      (in, game.layout_nb_columns * game.layout_nb_lines, game.pieces) ;
  if (game.layout == null) {
    in.close () ;
    return (null) ;
  }

  /* The image file data. First, we read it. */
  game.image_file_size = in.readInt () ;
  game.image_file_data = new byte [game.image_file_size] ;
  in.read (game.image_file_data, 0, game.image_file_size) ;

  /* Now, rebuild a CGImage ref from this content. */
  game.game_image =
    rebuildImageFromSavedBytes
      (game.image_file_data, game.image_file_size,
       game.original_image_scale_factor, game.rotated_image) ;
  if (game.game_image == null) return (null) ;

  /* We must now texture the pieces. */
  nb_columns_of_pieces = game.image_width / game.piece_width ;
  nb_lines_of_pieces = game.image_height / game.piece_height ;
  for (x = 0; x < nb_columns_of_pieces; x++) {
    for (y = 0; y < nb_lines_of_pieces; y++) {
      Point initial_center_position = new Point () ;
    
      /* Since the game was loaded, pieces moved compared to their position in a
         new game before shuffling. So, rebuild the position of the center. */
      initial_center_position.x =
        x * game.piece_width + (game.piece_width / 2) ;
      initial_center_position.y =
        y * game.piece_height + (game.piece_height / 2) ;
      if (! TextureMap.initializePieceTexture
              (game.game_image, game.pieces[(y * nb_columns_of_pieces) + x],
               initial_center_position)) {
        in.close () ;
        return (null) ;
      }
    }    /* End of for (y = 0; y < nb_lines_of_pieces; y++). */
  }      /* End of for (x = 0; x < game->nb_pieces; x++). */


  in.close () ;
  return (game) ;
  }

  
  
  static public void deleteSavegameFile (String file_name)
  {
    File sdCard = Environment.getExternalStorageDirectory () ; 
    File in_file = new File (sdCard, file_name) ;
    if (in_file.exists ()) in_file.delete () ;
  }
}
