package francois.pessaux.androjig.android;

import francois.pessaux.androjig.android.GameStruct.Game ;
import francois.pessaux.androjig.android.Maths2d.Point ;
import francois.pessaux.androjig.android.PieceStruct.Edge ;
import francois.pessaux.androjig.android.PieceStruct.Edges_list ;
import francois.pessaux.androjig.android.PieceStruct.Piece ;
import francois.pessaux.androjig.android.PieceStruct.Pieces_list ;

public class GameManips {

  /** Tests if a point is on the left side of an oriented segment (or on this
      segment). */
  private static boolean isPointOnRightOfSegment (Point pt,
                                                  Point seg_start,
                                                  Point seg_end)
  {
    if (((seg_start.x - seg_end.x) * (pt.y - seg_end.y)) <=
        ((seg_start.y - seg_end.y) * (pt.x - seg_end.x)))
      return (true) ;
    else return (false) ;
  }



  /** Test if a point in the interior of a piece. This is done by testing that
      the point is on the left side of each segment forming the bounding box
      (oriented N-E-S-W) of the piece. */
  private static boolean isPointInPieceBoundingBox (Point pt, Piece piece)
  {
    /* Recover the coordinates of the points of the bounding according to the
       current piece's center position. */
    int top_left_x = piece.center.x + piece.delta_center_to_top_left.x ;
    int top_left_y = piece.center.y + piece.delta_center_to_top_left.y ;
    int bottom_right_x = piece.center.x + piece.delta_center_to_bottom_right.x ;
    int bottom_right_y = piece.center.y + piece.delta_center_to_bottom_right.y ;
    Point seg_pt1 = new Point () ;
    Point seg_pt2 = new Point () ;
   
    seg_pt1.x = top_left_x ;
    seg_pt1.y = top_left_y ;
    seg_pt2.x = bottom_right_x ;
    seg_pt2.y = top_left_y ;
    if (! isPointOnRightOfSegment (pt, seg_pt1, seg_pt2)) return (false) ;
    /* Keep seg_pt2 as starting point because it already contains the right
       coordinates. */
    seg_pt1.x = bottom_right_x ;
    seg_pt1.y = bottom_right_y ;
    if (! isPointOnRightOfSegment (pt, seg_pt2, seg_pt1)) return (false) ;
    /* Keep seg_pt1 as starting point because it already contains the right
       coordinates. */
    seg_pt2.x = top_left_x ;
    seg_pt2.y = bottom_right_y ;
     if (! isPointOnRightOfSegment (pt, seg_pt1, seg_pt2)) return (false) ;
    /* Keep seg_pt2 as starting point because it already contains the right
       coordinates. */
    seg_pt1.x = top_left_x ;
    seg_pt1.y = top_left_y ;
    if (! isPointOnRightOfSegment (pt, seg_pt2, seg_pt1)) return (false) ;
    return (true) ;
  }



  /** Finds the piece whose center is the closest from the onscreen
      x and y coordinates. Returns this piece. If several pieces are
      at the same smallest distance from (x, y), then one of them is
      returned without special choice.

     - game : The current game structure.
     - x, y : Onscreen coordinates of the point near of which we want
       to find a piece. */
  public static Piece findClosestPiece (Game game, int x, int y)
  {
    /* Determine the piece's line/column position in the layout from
       the piece's center onscreen coordinates. */
    int x_in_layout = x / game.piece_width ;
    int y_in_layout = y / game.piece_height ;

    if ((x_in_layout < 0) || (x_in_layout >= game.layout_nb_columns) ||
        (y_in_layout < 0) || (y_in_layout >= game.layout_nb_lines)) {
      return (null) ; /* The point is out of the board. */
    }
    else {
      /* OK, the point is inside the board area. */
      Point mouse_point = new Point () ;
      Piece closest = null ;
      float min_dist = (float) 65536.0 ;

      mouse_point.x = x ;
      mouse_point.y = y ;
      /* Reduce the search space to only pieces that span on the layout's cell
         corresponding to the coordinates we received. */
      if ((x_in_layout >= 0) && (x_in_layout < game.layout_nb_columns) &&
          (y_in_layout >= 0) && (y_in_layout < game.layout_nb_lines)) {
        /* Compute the index in the layout corresponding to the selection. */
        int i = (y_in_layout * game.layout_nb_columns) + x_in_layout ;

        Pieces_list candidates = game.layout[i].hosted_pieces ;
        while (candidates != null) {
         Piece piece = candidates.piece ;
         
         /* Ensure that the point clicked is inside the piece's bouding-box. */
         if (isPointInPieceBoundingBox (mouse_point, piece)) {
           /* If the point is inside the piece's bounding-box, remind the piece
              whose center is the closest from the clicked point. */
           float dist = Maths2d.pointsDist (piece.center, mouse_point) ;
           if (dist < min_dist) {
             min_dist = dist ;
             closest = piece ;
           }   /* End of if (dist < min_dist). */
         }     /* End of if (isPointInPieceBoundingBox (&mouse_point, piece)). */
         candidates = candidates.next ;
       }
     }
     return (closest) ;
   }
 }



  private static void removePieceFromLayout (Game game, Piece piece)
  {
    int x, y ;
    int top_left_cell_x =
      (piece.center.x + piece.delta_center_to_top_left.x) /
      game.piece_width ;
    int top_left_cell_y =
      (piece.center.y + piece.delta_center_to_top_left.y) /
      game.piece_height ;
    int bottom_right_cell_x =
      (piece.center.x + piece.delta_center_to_bottom_right.x) /
      game.piece_width ;
    int bottom_right_cell_y =
      (piece.center.y + piece.delta_center_to_bottom_right.y) /
      game.piece_height ;
           
    for (x = top_left_cell_x; x <= bottom_right_cell_x; x++) {
      for (y = top_left_cell_y; y <= bottom_right_cell_y; y++) {
        if ((x >= 0) && (x < game.layout_nb_columns) &&
            (y >= 0) && (y < game.layout_nb_lines)) {
          /* Compute the index in the layout corresponding to the selection. */
          int i = (y * game.layout_nb_columns) + x ;

          /* We must now remove the piece [piece] from the list. */
          Pieces_list curr = game.layout[i].hosted_pieces ;
          Pieces_list prev = null ;
          while (curr != null) {
            if (curr.piece == piece) {
              /* We found the piece to remove. */
              if (prev == null) {
                /* The piece was the first of the list. */
                game.layout[i].hosted_pieces = curr.next ;
              }      /* if (prev == null). */
              else {
                /* The piece was not the first of the list. */
                prev.next = curr.next ;
              }
              curr = curr.next ;
            }     /* End of if (curr.piece == piece). */
            else {
              prev = curr ;
              curr = curr.next ;
            }
          }     /* End of while (curr != null). */
        }
      }         /* End of for (y = top_left_cell_y...). */
    }           /* End of for (x = top_left_cell_x...). */
  }



  /** Return FALSE in case of failure (memory missing). */
  public static boolean insertPieceInLayout (Game game, Piece piece)
  {
    int x, y ;
    int top_left_cell_x =
      (piece.center.x + piece.delta_center_to_top_left.x) /
      game.piece_width ;
    int top_left_cell_y =
      (piece.center.y + piece.delta_center_to_top_left.y) /
      game.piece_height ;
    int bottom_right_cell_x =
      (piece.center.x + piece.delta_center_to_bottom_right.x) /
      game.piece_width ;
    int bottom_right_cell_y =
      (piece.center.y + piece.delta_center_to_bottom_right.y) /
      game.piece_height ;

    for (x = top_left_cell_x; x <= bottom_right_cell_x; x++) {
      for (y = top_left_cell_y; y <= bottom_right_cell_y; y++) {
        if ((x >= 0) && (x < game.layout_nb_columns) &&
            (y >= 0) && (y < game.layout_nb_lines)) {
          /* Compute the index in the layout corresponding to the selection. */
          int i = (y * game.layout_nb_columns) + x ;

          Pieces_list new_head = new Pieces_list () ;
          /* Insert the piece in head of the list. */
          new_head.piece = piece ;
          new_head.next = game.layout[i].hosted_pieces ;
          game.layout[i].hosted_pieces = new_head ;
        }
      }
    }

    return (true) ;
  }



  private static void detachEdgeFromXSortedList (Game game, Edge edge)
  {
    if (edge.x_sorted_prev == null) {
      /* Piece is the first one, so update the head of the list. */
      game.x_sorted_edges_head = edge.x_sorted_next ;
    }
    else {
      edge.x_sorted_prev.x_sorted_next = edge.x_sorted_next ;
    }
    if (edge.x_sorted_next == null) {
      /* Piece is the last one, so update the tail of the list. */
      game.x_sorted_edges_tail = edge.x_sorted_prev ;
    }
    else {
      edge.x_sorted_next.x_sorted_prev = edge.x_sorted_prev ;
    }
    edge.x_sorted_prev = null ;
    edge.x_sorted_next = null ;
  }



  /** Moves the edge in the increasing sorted double-linked list,
     knowing that insertion will be done for an x lower that the
     initial one. This moves the edge's pointer "to the left" of
     the list until the sorting order is fulfilled.
     Attention, the piece must have its points' coordinates already
     updated to the new position. */
  private static void updateDecreasedXInList (Game game, Edge edge)
  {
    Edge insert_after ;
    int edge_min_x ;
    boolean go_on ;

    /* Remind from where we must search backward before detaching the
      current piece. */
    insert_after = edge.x_sorted_prev ;
    /* We first detach the current edge from the list and attach its 2
      neighbours. */
    detachEdgeFromXSortedList (game, edge) ;
    /* Now, walk backward until we find an edge with a minimal x <= to
      our current edge's minimal x coordinate in order to insert the
      current edge just after. */
    edge_min_x = Math.min (edge.p0.x, edge.p1.x) ;
    go_on = true ;
    while (go_on) {
      if (insert_after == null) go_on = false ;
      else {
        int ia_min_x = Math.min (insert_after.p0.x, insert_after.p1.x) ;
        if (ia_min_x <= edge_min_x) go_on = false ;
        else insert_after = insert_after.x_sorted_prev ;
      }
    }
    /* Now, insert after the found edge, or in head if the found edge
      doesn't exist because the current edge must be in fact the first
      one of the list. */
    if (insert_after == null) {
      edge.x_sorted_next = game.x_sorted_edges_head ;
      edge.x_sorted_prev = null ;
      if (game.x_sorted_edges_head != null)
        game.x_sorted_edges_head.x_sorted_prev = edge ;
      game.x_sorted_edges_head = edge ;
    }
    else {
      edge.x_sorted_next = insert_after.x_sorted_next ;
      if (insert_after.x_sorted_next == null)
        game.x_sorted_edges_tail = edge ;
      else
        insert_after.x_sorted_next.x_sorted_prev = edge ;
      insert_after.x_sorted_next = edge ;
      edge.x_sorted_prev = insert_after ;
    }
  }



  /** Moves the edge in the increasing sorted double-linked list,
     knowing that insertion will be done for an x greater that the
     initial one. This moves the edge's pointer "to the right" of
     the list until the sorting order is fulfilled.
     Attention, the piece must have its points' coordinates already
     updated to the new position. */
  private static void updateIncreasedXInList (Game game, Edge edge)
  {
    Edge insert_before ;
    int edge_min_x ;
    boolean go_on ;

    /* Remind from where we must search forward before detaching the
      current piece. */
    insert_before = edge.x_sorted_next ;
    /* We first detach the current edge from the list and attach its 2
      neighbours. */
    detachEdgeFromXSortedList (game, edge) ;
    /* Now, walk forward until we find an edge with a minimal x >= to
      our current edge's minimal x coordinate in order to insert the
      current edge just before. */
    edge_min_x = Math.min (edge.p0.x, edge.p1.x) ;
    go_on = true ;
    while (go_on) {
      if (insert_before == null) go_on = false ;
      else {
        int ib_min_x = Math.min (insert_before.p0.x, insert_before.p1.x) ;
        if (ib_min_x >= edge_min_x) go_on = false ;
        else insert_before = insert_before.x_sorted_next ;
      }
    }
    /* Now, insert before the found edge, or in tail if the found edge
      doesn't exist because the current edge must be in fact the last
      one of the list. */
    if (insert_before == null) {
      edge.x_sorted_next = null ;
      edge.x_sorted_prev = game.x_sorted_edges_tail ;
      if (game.x_sorted_edges_tail != null)
        game.x_sorted_edges_tail.x_sorted_next = edge ;
      game.x_sorted_edges_tail = edge ;
    }
    else {
      edge.x_sorted_next = insert_before ;
      if (insert_before.x_sorted_prev == null)
        game.x_sorted_edges_head = edge ;
      else insert_before.x_sorted_prev.x_sorted_next = edge ;
      edge.x_sorted_prev = insert_before.x_sorted_prev ;
      insert_before.x_sorted_prev = edge ;
    }
  }



  private static void detachEdgeFromYSortedList (Game game, Edge edge)
  {
    if (edge.y_sorted_prev == null) {
      /* Piece is the first one, so update the head of the list. */
      game.y_sorted_edges_head = edge.y_sorted_next ;
    }
    else {
      edge.y_sorted_prev.y_sorted_next = edge.y_sorted_next ;
    }
    if (edge.y_sorted_next == null) {
      /* Piece is the last one, so update the tail of the list. */
      game.y_sorted_edges_tail = edge.y_sorted_prev ;
    }
    else {
      edge.y_sorted_next.y_sorted_prev = edge.y_sorted_prev ;
    }
    edge.y_sorted_prev = null ;
    edge.y_sorted_next = null ;
  }



  /** Moves the edge in the increasing sorted double-linked list,
     knowing that insertion will be done for an y lower that the
     initial one. This moves the edge's pointer "to the left" of
     the list until the sorting order is fulfilled.
     Attention, the piece must have its points' coordinates already
     updated to the new position. */
  private static void updateDecreasedYInList (Game game, Edge edge)
  {
    Edge insert_after ;
    int edge_min_y ;
    boolean go_on ;

    /* Remind from where we must search backward before detaching the
      current piece. */
    insert_after = edge.y_sorted_prev ;
    /* We first detach the current edge from the list and attach its 2
      neighbours. */
    detachEdgeFromYSortedList (game, edge) ;
    /* Now, walk backward until we find an edge with a minimal y <= to
     our current edge's minimal y coordinate in order to insert the
     current edge just after. */
    edge_min_y = Math.min (edge.p0.y, edge.p1.y) ;
    go_on = true ;
    while (go_on) {
      if (insert_after == null) go_on = false ;
      else {
        int ia_min_y = Math.min (insert_after.p0.y, insert_after.p1.y) ;
        if (ia_min_y <= edge_min_y) go_on = false ;
        else insert_after = insert_after.y_sorted_prev ;
      }
    }
    /* Now, insert after the found edge, or in head if the found edge
      doesn't exist because the current edge must be in fact the first
     one of the list. */
    if (insert_after == null) {
      edge.y_sorted_next = game.y_sorted_edges_head ;
      edge.y_sorted_prev = null ;
      if (game.y_sorted_edges_head != null)
        game.y_sorted_edges_head.y_sorted_prev = edge ;
      game.y_sorted_edges_head = edge ;
    }
    else {
      edge.y_sorted_next = insert_after.y_sorted_next ;
      if (insert_after.y_sorted_next == null)
        game.y_sorted_edges_tail = edge ;
      else insert_after.y_sorted_next.y_sorted_prev = edge ;
      insert_after.y_sorted_next = edge ;
      edge.y_sorted_prev = insert_after ;
    }
  }



  /** Moves the edge in the increasing sorted double-linked list,
     knowing that insertion will be done for an y greater that the
     initial one. This moves the edge's pointer "to the right" of
     the list until the sorting order is fulfilled.
     Attention, the piece must have its points' coordinates already
     updated to the new position. */
  private static void updateIncreasedYInList (Game game, Edge edge)
  {
    Edge insert_before ;
    int edge_min_y ;
    boolean go_on ;

    /* Remind from where we must search backward before detaching the
      current piece. */
    insert_before = edge.y_sorted_next ;
    /* We first detach the current edge from the list and attach its 2
      neighbours. */
    detachEdgeFromYSortedList (game, edge) ;
    /* Now, walk forward until we find an edge with a minimal y >= to
      our current edge's minimal y coordinate in order to insert the
      current edge just before. */
    edge_min_y = Math.min (edge.p0.y, edge.p1.y) ;
    go_on = true ;
    while (go_on) {
      if (insert_before == null) go_on = false ;
      else {
        int ib_min_y = Math.min (insert_before.p0.y, insert_before.p1.y) ;
        if (ib_min_y >= edge_min_y) go_on = false ;
        else insert_before = insert_before.y_sorted_next ;
      }
    }
    /* Now, insert before the found edge, or in tail if the found edge
      doesn't exist because the current edge must be in fact the last
      one of the list. */
    if (insert_before == null) {
      edge.y_sorted_next = null ;
      edge.y_sorted_prev = game.y_sorted_edges_tail ;
      if (game.y_sorted_edges_tail != null)
        game.y_sorted_edges_tail.y_sorted_next = edge ;
      game.y_sorted_edges_tail = edge ;
    }
    else {
      edge.y_sorted_next = insert_before ;
      if (insert_before.y_sorted_prev == null)
        game.y_sorted_edges_head = edge ;
      else insert_before.y_sorted_prev.y_sorted_next = edge ;
      edge.y_sorted_prev = insert_before.y_sorted_prev ;
      insert_before.y_sorted_prev = edge ;
    }
  }



  /** Local function that performs a shift of only one piece. It will be applied
     on each piece belonging to the connected-set of the initially selected
     piece. */
  private static void shiftOnePiece (int dx, int dy, Game game, Piece piece)
  {
    int i ;

    /* Remove the piece from its old position in the layout. Do this
      BEFORE the piece's points are shifted to the new position. */
    removePieceFromLayout (game, piece) ;

    /* Update the center's position. */
    piece.center.x = piece.center.x + dx ;
    piece.center.y = piece.center.y + dy ;

    /* Update the splines points of the four sides of the piece and shift
      edges extremities of the piece. */
    for (i = 0; i < 4; i++) {
      Edge edge = piece.free_edges[i] ;

      /* Since we do not draw the splines, and since they are only used to
        initialize the texture mapping, we don't need to move the splines'
        points. This saves computation.
        Note that if we want to draw the splines, then they must be moved with
        the code below,
     for (j = 0; j < edge->spline->nb_points; j++) {
       SHIFT_POINT (edge->spline->vertices[j]) ;
     }
        and in this case, when loading a game, the center position of the piece
        that must be used to initialize texture must not be anymore the *initial*
        position, but the current position of the center of the piece, i.e.
        with coordinates changed by the movements applied to the piece !
        Moreover in this case, the splines' points must be moved even if the
        edge is no more active otherwise edges marked as inactive would not have
        their splines points move consistently with the center of the piece and
        trying to rebuild the polygon forming the shape of the piece would take
        wrong points coordinates. */

      if (edge.active) {
        /* Shift the edge extremities. */
        edge.p0.x = edge.p0.x + dx ;
        edge.p0.y = edge.p0.y + dy ;
        edge.p1.x = edge.p1.x + dx ;
        edge.p1.y = edge.p1.y + dy ;

        /* Update the position of the edges of the piece in the x / y
          increasing sorted double-linked lists. Do this here and not
          after all the edges have moved because if we update before all the
          edges points, then while updating afterwards in the list, edges will
          already have their points positions modified while not yet well
          sorted in the list. So when we would try to find where to insert an
          edge, may be we will find in the list an edge already
          position-updated but not-sort updated. Hence we would insert edges
          relatively to not well-sorted edges. */
        if (dx < 0)
          updateDecreasedXInList (game, edge) ; /* Piece decreased in x. */
        else updateIncreasedXInList (game, edge) ; /* Piece decreased in x. */
        if (dy < 0)
          updateDecreasedYInList (game, edge) ; /* Piece decreased in y. */
        else updateIncreasedYInList (game, edge) ; /* Piece decreased in y. */
      }    /* End of if (edge->active). */
    }      /* End of for (i = 0; i < 4; i++). */

    /* Now, also move the piece's location in term of line/column in
      the layout: add it to its new position in the layout. For this, we don't
      use the edges coordinate, and that's great since we shifted them only if
      they were still active. We use the piece's center position and the bounding
      box. And of course, the center is always shifted ! */
    insertPieceInLayout (game, piece) ;
  }



  /** Moves a piece to a given position. All the connected pieces
     (i.e. those belonging to the same connected-set that this piece)
     will then follow the same displacement.
     Basically, this function updates all the pieces' points'
     coordinates (pieces' center point, edges' points and splines
     points) and moves each piece in the double-linked lists sorted by
     increasing x / y in order to keep these lists consistently sorted
     with the new pieces' positions. It also moves the piece in the
     layout. */
  public static void movePieceTo (int x, int y, Game game, Piece hooked_piece)
  {
    Pieces_list list_cell ;
    /* Compute the delta representing the piece's movement. */
    int dx = x - hooked_piece.center.x ;
    int dy = y - hooked_piece.center.y ;

    /* Now, do the job and really shift all the pieces member of the
      same connected-set (this set at least contains the current
      piece). */
    list_cell = hooked_piece.connected_pieces.members ;
    while (list_cell != null) {
      shiftOnePiece (dx, dy, game, list_cell.piece) ;
      list_cell = list_cell.next ;
    }
  }



  private static final int COALESCENCE_TOLERANCE = 12 ;



  private static Edges_list walkBackwardX (Edges_list cons_to, Edge edge)
  {
    Edge e ;
    int original_min_x, original_max_x ;

    if (edge == null) return (cons_to) ;
    e = edge ;
    original_min_x = Math.min (edge.p0.x, edge.p1.x) ;
    original_max_x = Math.max (edge.p0.x, edge.p1.x) ;

    while (e != null) {
      int e_min_x = Math.min (e.p0.x, e.p1.x) ;
      int e_max_x = Math.max (e.p0.x, e.p1.x) ;
      /* Note that before inserting e in the list, we could check that
        its y coordinates fit the original's piece y coordinates modulo
        the tolerance but in any case, this will be done when we will
        check that anchor points match. This would only duplicate the
        test to save a bit of memory allocation (cons in the list). */
      if (((Math.abs (original_min_x - e_min_x)) <= COALESCENCE_TOLERANCE) ||
          ((Math.abs (original_max_x - e_max_x)) <= COALESCENCE_TOLERANCE)) {
        Edges_list new_element = new Edges_list () ;
        new_element.edge = e ;
        new_element.next = cons_to ;
        cons_to = new_element ;
        e = e.x_sorted_prev ;
      }
      else e = null ;  /* Stop search. */
    }
    return (cons_to) ;
  }




  static private Edges_list walkBackwardY (Edges_list cons_to, Edge edge)
  {
    Edge e ;
    int original_min_y, original_max_y ;

    if (edge == null) return (cons_to) ;
    e = edge ;
    original_min_y = Math.min (edge.p0.y, edge.p1.y) ;
    original_max_y = Math.max (edge.p0.y, edge.p1.y) ;

    while (e != null) {
      int e_min_y = Math.min (e.p0.y, e.p1.y) ;
      int e_max_y = Math.max (e.p0.y, e.p1.y) ;
      if (((Math.abs (original_min_y - e_min_y)) <= COALESCENCE_TOLERANCE) ||
          ((Math.abs (original_max_y - e_max_y)) <= COALESCENCE_TOLERANCE)) {
        Edges_list new_element = new Edges_list () ;
        new_element.edge = e ;
        new_element.next = cons_to ;
        cons_to = new_element ;
        e = e.y_sorted_prev ;
      }
      else e = null ;  /* Stop search. */
    }
    return (cons_to) ;
  }



  private static Edges_list walkForwardX (Edges_list cons_to, Edge edge)
  {
    Edge e ;
    int original_min_x, original_max_x ;

    if (edge == null) return (cons_to) ;
    e = edge ;
    original_min_x = Math.min (edge.p0.x, edge.p1.x) ;
    original_max_x = Math.max (edge.p0.x, edge.p1.x) ;

    while (e != null) {
      int e_min_x = Math.min (e.p0.x, e.p1.x) ;
      int e_max_x = Math.max (e.p0.x, e.p1.x) ;
      if (((Math.abs (original_min_x - e_min_x)) <= COALESCENCE_TOLERANCE) ||
          ((Math.abs (original_max_x - e_max_x)) <= COALESCENCE_TOLERANCE)) {
        Edges_list new_element = new Edges_list () ;
        new_element.edge = e ;
        new_element.next = cons_to ;
        cons_to = new_element ;
        e = e.x_sorted_next ;
      }
      else e = null ;  /* Stop search. */
    }
    return (cons_to) ;
  }



  private static Edges_list walkForwardY (Edges_list cons_to, Edge edge)
  {
    Edge e ;
    int original_min_y, original_max_y ;

    if (edge == null) return (cons_to) ;
    e = edge ;
    original_min_y = Math.min (edge.p0.y, edge.p1.y) ;
    original_max_y = Math.max (edge.p0.y, edge.p1.y) ;

    while (e != null) {
      int e_min_y = Math.min (e.p0.y, e.p1.y) ;
      int e_max_y = Math.max (e.p0.y, e.p1.y) ;
      if (((Math.abs (original_min_y - e_min_y)) <= COALESCENCE_TOLERANCE) ||
          ((Math.abs (original_max_y - e_max_y)) <= COALESCENCE_TOLERANCE)) {
        Edges_list new_element = new Edges_list () ;
        new_element.edge = e ;
        new_element.next = cons_to ;
        cons_to = new_element ;
        e = e.y_sorted_next ;
      }
      else e = null ;  /* Stop search. */
    }
    return (cons_to) ;
  }



  private static Edges_list findCloseEdges (Edge edge)
  {
    Edges_list backward_x_edges ;
    Edges_list backward_y_edges ;
    Edges_list forward_x_edges ;
    Edges_list all_edges ;
    Edges_list curr ;
    Edges_list result_head ;

    backward_x_edges = walkBackwardX (null, edge.x_sorted_prev) ;
    backward_y_edges = walkBackwardY (backward_x_edges, edge.y_sorted_prev) ;
    forward_x_edges = walkForwardX (backward_y_edges, edge.x_sorted_next) ;
    all_edges = walkForwardY (forward_x_edges, edge.y_sorted_next) ;

    /* Since an edge can appear twice because of searches in both x and
      y lists, we ensure to make edges appear only once in the result
      list. */
    curr = all_edges ;
    result_head = null ;
    while (curr != null) {
      Edges_list walker = curr.next ;
      boolean found_double = false ;

      /* Search if we find the current edge among the remaining ones. */
      while (walker != null) {
        if (curr.edge == walker.edge) {
          walker = null ;    /* Stop search. */
          found_double = true ;
        }
        else walker = walker.next ;
      }   /* End of while (walker != null). */

      /* Check the boolean telling if we found the edge twice. */
      if (found_double) {
        curr = curr.next ;
      }        /* End of if (found_double). */
      else {
        Edges_list tmp = curr.next ;
        /* Keep the current edge in the result list. */
        curr.next = result_head ;
        result_head = curr ;
        curr = tmp ;
      }        /* End of else. */
    }          /* End of while (curr != NULL). */

    return (result_head) ;
  }



  /** [Descr] : Joins the 2 pieces owning the edges passed as
     argument. Hence, these 2 pieces are getting belonging to the same
     connected-set. Since membership is based on physical sharing of
     the connected-set structure, we simple add p2's connected-set
     members to p1's connected-set and make p2 sharing this
     connected-set.
     It can arrive that we are requested to coalesce 2 edges already
     belonging to the same connected-set. This is the case when 2
     pieces coalesce by several edges. The 2 first edges make the
     pieces coalescing first, hence puttign them into the same
     connected-set. Then the other edges are examined, seen as also
     coalescing. In this case, we don't need to modify the
     connected-set, we just need to destroy the edges stuff.
     [Rem] : Not exported outside this module. */
  private static void doPiecesCoalesceForEdges (Game game, Edge edge1,
      Edge edge2)
  {
    int i ;
    Piece piece1, piece2 ;

    piece1 = edge1.part_of_piece ;
    piece2 = edge2.part_of_piece ;

    /* Merge both pieces into the same connected-set if it is not already
      the case. */
    if (piece1.connected_pieces != piece2.connected_pieces) {
      Pieces_list last_of_p1, tmp ;

      /* First, search the last element of piece1's members list to make
        later its field [next] pointing to the first element of piece2's
        members list. Hence the 2 lists will become linked into one. */
      last_of_p1 = piece1.connected_pieces.members ;
      while ((last_of_p1 != null) && (last_of_p1.next != null))
        last_of_p1 = last_of_p1.next ;
      /* We keep the piece1's connected-set as the one shared. So we
        must add piece2's members inside. Hence, we simply extend
        piece1's list by piece2's list. */
      if (last_of_p1 != null)
        last_of_p1.next = piece2.connected_pieces.members ;
      /* Now, for each member of piece2, we must make it sharing
        piece1's updated connected-set. Since piece2 is in it
        connected-set, iterating on piece2's members will also
        update piece2. */
      tmp = piece2.connected_pieces.members ;
      while (tmp != null) {
        tmp.piece.connected_pieces = piece1.connected_pieces ;
        /* Update the reference counter of the shared connected-set. */
        piece1.connected_pieces.refcount++ ;
        tmp = tmp.next ;
      }
      /* MEMLEAK: piece2->connected_pieces (i.e. connected-set
        structure, not its inner list) could normally be released. */
    }

    /* Remove both edges from the x / y sorted lists. */
    detachEdgeFromXSortedList (game, edge1) ;
    detachEdgeFromXSortedList (game, edge2) ;
    detachEdgeFromYSortedList (game, edge1) ;
    detachEdgeFromYSortedList (game, edge2) ;
    /* "Remove" the edges from their respective pieces (mark them not
      active). */
    for (i = 0; i < 4; i++) {
      if (piece1.free_edges[i] == edge1) piece1.free_edges[i].active = false ;
    }
    for (i = 0; i < 4; i++) {
      if (piece2.free_edges[i] == edge2) piece2.free_edges[i].active = false ;
    }
  }


  
  /** [Descr] : Checks if the piece passed as argument can coalesce with
     another one in the current game configuration. This function only
     processes the piece passed as argument and doesn't apply to the
     other pieces of the same connected-set than the one passed as
     argument.
     [Rem] : Not exported] outside this module. */
  private static boolean tryToCoalesceOnePiece (Game game, Piece piece)
  {
    int edge_index ;
    boolean foundCoalescing = false ;

    /* Hunt all pieces around the current one that are near of it, at
      most at a distance equal to the coalescence tolerance. We must
      hence scan edges near of the current piece's 4 edges. */
    for (edge_index = 0; edge_index < 4; edge_index++) {
      Edge edge = piece.free_edges[edge_index] ;
      if (edge.active) {
        Edges_list close_edges = findCloseEdges (edge) ;
        while (close_edges != null) {
          Edge close_edge = close_edges.edge ;
          /* Check if the found close edge is the one matching the
            initial edge. */
          if (((Math.abs (edge.p0.x - close_edge.p0.x)) <= COALESCENCE_TOLERANCE) &&
              ((Math.abs (edge.p0.y - close_edge.p0.y)) <= COALESCENCE_TOLERANCE)) {
            /* Points p0 of the 2 edges are close, so it is the case
              for points p1. */
            if ((edge.p0_anchor_id == close_edge.p0_anchor_id) &&
                (edge.p1_anchor_id == close_edge.p1_anchor_id)) {
              int x, y ;
              /* Compensate the tolerance, making the piece exactly fitting the
                one it matches. Do this before coalescence to prevent moving the
                newly merged set and displace pieces that do not need to move
                because we correct the position especially by taking their
                position as the right one. */
              x = piece.center.x + (close_edge.p0.x - edge.p0.x) ;
              y = piece.center.y + (close_edge.p0.y - edge.p0.y) ;
              movePieceTo (x, y, game, piece) ;
              doPiecesCoalesceForEdges (game, edge, close_edge) ;
              foundCoalescing = true ;
            }
          }
          else {
            if (((Math.abs (edge.p0.x - close_edge.p1.x)) <= COALESCENCE_TOLERANCE) &&
                ((Math.abs (edge.p0.y - close_edge.p1.y)) <= COALESCENCE_TOLERANCE)) {
              /* Points p0 and p1 of edges are close, so it is the
                case for points p1 and p0. */
              if ((edge.p0_anchor_id == close_edge.p1_anchor_id) &&
                  (edge.p1_anchor_id == close_edge.p0_anchor_id)) {
                int x, y ;
                /* Compensate the tolerance, making the piece exactly fitting the
                  one it matches. Do this before coalescence (same remark than
                  just above). */
                x = piece.center.x + (close_edge.p1.x - edge.p0.x) ;
                y = piece.center.y + (close_edge.p1.y - edge.p0.y) ;
                movePieceTo (x, y, game, piece) ;
                doPiecesCoalesceForEdges (game, edge, close_edge) ;
                foundCoalescing = true ;
              }
            }
          }
          /* Process the next edge of the close edges list. */
          close_edges = close_edges.next ;
        }

        /* Release the close_edges list. */
        PieceStruct.freeEdgesList (close_edges) ;
        close_edges = null ;
      }    /* End of if (edge->active). */
    }      /* End of for (edge_index = 0; edge_index < 4; edge_index++). */
    return (foundCoalescing) ;
  }



  /** [Descr] : Checks coalescence for all the pieces belonging to the
     same connected-set than the one passed as argument.
     Hence, this function is the entry point for piece coalescence
     computation in the game.
     [Rem] : Exported outside this module. */
  public static boolean checkCoalescence (Game game, Piece from_piece)
  {
    Pieces_list lst ;
    Pieces_list copy ;
    boolean foundCoalescing = false ;

    /* Because coalescance modifies in place the list of members of
    connected-sets, to prevent coalecences in chain, we make a copy of
    the connected-set of the original piece. */
    copy = PieceStruct.copyPiecesList (from_piece.connected_pieces.members) ;
    lst = copy ;

    /* Walk along the members of the connected set to make each of them
      coalescing with some neighbours. */
    while (lst != null) {
    	foundCoalescing =
    		foundCoalescing || tryToCoalesceOnePiece (game, lst.piece) ;
      lst = lst.next ;
    }

    /* Free the copy of the connected-set members. */
    PieceStruct.freePiecesList (copy) ;
    copy = null ;
    return (foundCoalescing) ;
  }



  public static void shuffleAllPieces (Game game)
  {
    int i ;
    int max_x = ((game.layout_nb_columns - 1) * game.piece_width) ;
    int max_y = ((game.layout_nb_lines - 1) * game.piece_height) ;

    for (i = 0; i < game.nb_pieces; i++) {
      int x, y ;
      Piece piece = game.pieces[i] ;

      /* Select a random position for the center of the piece. */
      x = AndroJigApp.randomGenerator.nextInt (max_x) ;
      y = AndroJigApp.randomGenerator.nextInt (max_y) ;

      /* Ensure that with its center to this position, the piece will not have
        parts outside the window. */
      if ((x + piece.delta_center_to_top_left.x) < 0)
        x = -piece.delta_center_to_top_left.x ;
      if (x + piece.delta_center_to_bottom_right.x >= max_x)
        x = max_x - piece.delta_center_to_bottom_right.x ;
      if ((y + piece.delta_center_to_top_left.y) < 0)
        y = -piece.delta_center_to_top_left.y ;
      if (y + piece.delta_center_to_bottom_right.y >= max_y)
        y = max_y - piece.delta_center_to_bottom_right.y ;

      /* Move the piece. */
      movePieceTo (x, y, game, piece) ;
    }
  }



  /** Shuffle pieces so that they are spreat along the border of the playfield,
     leaving the center region empty. */
  public static void arrangePiecesOnBorders (Game game)
  {
    int i, x, y, x_max, y_max, x_min, y_min, mode ;
    int piece_width, piece_height ;
    int[] indir ;

    /* Allocate the indirection array we will scan to send each piece at a
      random location. This array will contain the order in which the pieces
      will be taken to be sent at their new location. The piece whose index is
      in the first cell of the array will be moved first, the second will be
      moved in second, and so on. By randomizing this indirection array, we make
      so that is not always the same piece that is moved to each border
      position. */
    indir = new int [game.nb_pieces] ;

    /* Initialize the indirection array with identity. */
    for (i = 0; i < game.nb_pieces; i++) indir[i] = i ;

    /* Now, for each cell of the array, swap the index of the piece with a random
      index of another cell. */
    for (i = 0; i < game.nb_pieces; i++) {
      int tmp ;
      int rnd_index = AndroJigApp.randomGenerator.nextInt (game.nb_pieces) ;
      tmp = indir[i] ;
      indir[i] = indir[rnd_index] ;
      indir[rnd_index] = tmp ;
    }

    piece_width = game.piece_width ;
    piece_height = game.piece_height ;
    mode = PieceStruct.NORTH ;
    x_min = piece_width / 2 ;
    x_max = ((game.layout_nb_columns - 1) * piece_width) ;
    y_min = piece_height / 2 ;
    y_max = ((game.layout_nb_lines - 1) * piece_height) ;
    x = x_min ;
    y = y_min ;

    /* Walk along the indirection array, taking each piece whose id is in the
      array and send it to the next available position along the borders of the
      playfield. */
    for (i = 0; i < game.nb_pieces; i++) {
      Piece piece = game.pieces[indir[i]] ;

      /* Attention, we don't move pieces already coalesced with other ones. */
      if ((piece.connected_pieces.members != null) &&
          (piece.connected_pieces.members.next != null)) continue ;

      switch (mode) {
      case PieceStruct.NORTH:
        movePieceTo (x, y, game, piece) ;
        x += piece_width ;
        if (x > x_max) {
          mode = PieceStruct.EAST ;
          y += piece_height ;
          x = x_max ;
          x_max -= piece_width ;
        }
        break ;

      case PieceStruct.EAST:
        movePieceTo (x, y, game, piece) ;
        y += piece_height ;
        if (y > y_max) {
          mode = PieceStruct.SOUTH ;
          x -= piece_width ;
          y = y_max ;
          y_max -= piece_height ;
        }
        break ;

      case PieceStruct.SOUTH:
        movePieceTo (x, y, game, piece) ;
        x -= piece_width ;
        if (x < x_min) {
          mode = PieceStruct.WEST ;
          y -= piece_height ;
          x = x_min ;
          x_min += piece_width ;
        }
        break ;

      case PieceStruct.WEST:
        movePieceTo (x, y, game, piece) ;
        y -= piece_height ;
        if (y < y_min) {
          mode = PieceStruct.NORTH ;
          x += piece_width ;
          y = y_min ;
          y_min += piece_height ;
        }
        break ;
      }    /* End of switch (mode). */
    }      /* End of for (i = 0; i < game->nb_pieces; i++). */

  }



  /** Simply checks if the refcount of the connected-set of a piece of the game
     (the first one for instance) is equal to the number of pieces of the game.
     In effect, since when pieces colapse they share the same connected set and
     increment the related refcount, when all the pieces are connected, there
     remain only 1 connected set and its refcount must be equal to the number of
     pieces of the game. */
  public static boolean isGameJustWon (Game game)
  {
    if (game == null) return (false) ;
    /* Don't notify the game is won if it was already won. */
    if (game.game_won) return (false) ;
    if (game.nb_pieces <= 0) return (false) ;
    if (game.pieces[0].connected_pieces.refcount == game.nb_pieces) {
      game.game_won = true ;
      return (true) ;
    }
    else return (false) ;
  }

}
