package francois.pessaux.androjig.android;

import java.io.DataOutputStream;
import java.io.File ;
import java.io.FileOutputStream;
import java.io.IOException;

import android.os.Environment ;

import francois.pessaux.androjig.android.GameStruct.Game;
import francois.pessaux.androjig.android.GameStruct.Layout_cell;
import francois.pessaux.androjig.android.Maths2d.Point;
import francois.pessaux.androjig.android.PieceStruct.Connected_set;
import francois.pessaux.androjig.android.PieceStruct.Edge;
import francois.pessaux.androjig.android.PieceStruct.Piece;
import francois.pessaux.androjig.android.PieceStruct.Pieces_list;
import francois.pessaux.androjig.android.SplineStruct.Spline;

public class GameSave {
	
	private static void savePoint (DataOutputStream out, Point point)
		throws IOException
	{
		out.writeInt (point.x) ;
		out.writeInt (point.y) ;
	}
	
	
	
  private static void saveLayout (DataOutputStream out, Layout_cell[] layout,
  			 										      int nb_cells)
  	throws IOException
  {
    int i ;

    /* Save the list of pieces ids of each layout cell. Forget the redraw flag. */
    for (i = 0; i < nb_cells; i++) {
      Pieces_list list = layout[i].hosted_pieces ;
      int list_len ;

      list_len = PieceStruct.piecesListLength (list) ;
      /* Save the number of pieces in the list of the current cell. */
      out.writeInt (list_len) ;
      /* Now, save the id of each piece of the list. */
      while (list != null) {
      	out.writeInt (list.piece.piece_debug_id) ;
        list = list.next ;
      }    /* End of while (list != null). */
    }
  }



  private static void saveSpline (DataOutputStream out, Spline spline)
  	throws IOException
  {
    /* The number of points. */
    out.writeInt (spline.nb_points) ;
    /* The vertices themselves. */
    for (int i = 0; i < spline.nb_points; i++)
    	savePoint (out, spline.vertices[i]) ;
  }


  
  /** Saves all the edges present in the array passed as argument. 
   * @throws IOException */
  private static void saveAllEdges (DataOutputStream out, Edge[] edges,
  																	int nb_edges)
  	throws IOException
  {
    int tmp, i ;

    for (i = 0; i < nb_edges; i++) {
      Edge edge = edges[i] ;

      /* The id. */
      out.writeInt (edge.id) ;
      /* Activity. */
      out.writeBoolean (edge.active) ;
      /* Point p0. */
      savePoint (out, edge.p0) ;
      /* Point p0 anchor id. */
      out.writeInt (edge.p0_anchor_id) ;
      /* Point p1. */
      savePoint (out, edge.p1) ;
      /* Point p1 anchor id. */
      out.writeInt (edge.p1_anchor_id) ;
      /* The spline. */
      saveSpline (out, edge.spline) ;
      /* The id of the piece this edge belongs to. */
      out.writeInt (edge.part_of_piece.piece_debug_id) ;
      
      /* The ids of our next/prev edges in the X/Y sorted lists or -1 if NULL. */
      if (edge.x_sorted_prev == null) tmp = -1 ;
      else tmp = edge.x_sorted_prev.id ;
      out.writeInt (tmp) ;
      if (edge.x_sorted_next == null) tmp = -1 ;
      else tmp = edge.x_sorted_next.id ;
      out.writeInt (tmp) ;
      if (edge.y_sorted_prev == null) tmp = -1 ;
      else tmp = edge.y_sorted_prev.id ;
      out.writeInt (tmp) ;
      if (edge.y_sorted_next == null) tmp = -1 ;
      else tmp = edge.y_sorted_next.id ;
      out.writeInt (tmp) ;
    }
  }



  private static void savePieces (DataOutputStream out, Piece[] pieces,
  																int nb_pieces)
  	throws IOException
  {
    int i, j ;
    Connected_set[] found_connected_sets ;
    Edge[] all_edges ;
    int nb_found_connected_sets ;

    /* Allocate an array to remind the pointers of already seen connected-sets.
       In the worst case, we will have as many connected-sets as pieces, so to
       make simpler, directly allocate an array of the maximum size. */
    found_connected_sets = new Connected_set [nb_pieces] ;
    nb_found_connected_sets = 0 ;

    /* First, identify and count the connected-sets used by the piece. We walk
       along all the pieces and remind the connected-sets appearing we didn't
       already saw. We store them in the array found_connected_sets. */
    for (i = 0; i < nb_pieces; i++) {
      boolean already_seen ;
      Connected_set current_cs = pieces[i].connected_pieces ;

      already_seen = false ;
      for (j = 0; j < nb_found_connected_sets; j++) {
        if (current_cs == found_connected_sets[j]) {
          /* The current connected-set already belongs to the seen ones. */
          already_seen = true ;
          break ;
        }    /* End of if (current_cs == found_connected_sets[j]). */
      }      /* End of for (j = 0; j < nb_found_connected_sets; j++). */

      /* Check if we exited without finding the current connected-set in the
         already seen ones. If so, add it to the seen ones. */
      if (already_seen != true) {
        found_connected_sets[nb_found_connected_sets] = current_cs ;
        nb_found_connected_sets++ ;
      }      /* End of if (already_seen != true). */
    }        /* End of for (i = 0; i < nb_pieces; i++). */

    /* Now, save the found unique connected-sets, starting by the number we
       found of. */
    out.writeInt (nb_found_connected_sets) ;

    for (i = 0; i < nb_found_connected_sets; i++) {
      Connected_set current_cs = found_connected_sets[i] ;
      Pieces_list plist = current_cs.members ;
      /* Count the number of pieces in this connected-set and save it. This should
         be directly the refcount, but... it doesn't hurt. */
      int len = PieceStruct.piecesListLength (plist) ;
      out.writeInt (len) ;

      /* Now, save the refcount. */
      out.writeInt (current_cs.refcount) ;

      /* Now, save the id of each piece member of this connected-set. */
      while (plist != null) {
      	out.writeInt (plist.piece.piece_debug_id) ;
        plist = plist.next ;
      }    /* End of while (plist != null). */
    }      /* End of for (i = 0; i < nb_found_connected_sets; i++). */

    /* Now, save all the edges present in all the pieces. We first make an array
       containing all of them. */
    all_edges = new Edge [nb_pieces * 4] ;
    
    /* Use j to count edges we store in the array. Walk along all the pieces,
       harvesting the pointers of their edges. */
    j = 0 ;
    for (i = 0; i < nb_pieces; i++) {
      Piece piece = pieces[i] ;
      all_edges[j] = piece.free_edges[PieceStruct.NORTH] ;
      j++ ;
      all_edges[j] = piece.free_edges[PieceStruct.EAST] ;
      j++ ;
      all_edges[j] = piece.free_edges[PieceStruct.SOUTH] ;
      j++ ;
      all_edges[j] = piece.free_edges[PieceStruct.WEST] ;
      j++ ;
    }
    
    /* Now, really write the edges information. */
    saveAllEdges (out, all_edges, nb_pieces * 4) ;
    all_edges = null ;

    
    /* Now, really save the pieces structures. Note that the image stuff is not
       saved, they will be rebuilt at loading-time. */
    for (i = 0; i < nb_pieces; i++) {
      Piece piece = pieces[i] ;

      /* The id. */
      out.writeInt (piece.piece_debug_id) ;
      /* The center. */
      savePoint (out, piece.center) ;
      /* The bounding-box. */
      savePoint (out, piece.delta_center_to_top_left) ;
      savePoint (out, piece.delta_center_to_bottom_right) ;
      /* The edges ids. */
      for (j = 0; j < 4; j++) out.writeInt (piece.free_edges[j].id) ;

      /* The connected-set. We save the index of the connected set we previously
         saved that corresponds to the current piece's connected-set. Hence, we
         must recover it from the array by comparing its pointer with those stored
         in this array. */
      for (j = 0; j < nb_found_connected_sets; j++) {
        if (found_connected_sets[j] == piece.connected_pieces) {
          /* We found it, save its index in the file. */
        	out.writeInt (j) ;
          break ;
        }
      }    /* End of for (j = 0; j < nb_found_connected_sets; j++). */

      /* For sake of safety, ensure we really found the searched connected-set.
         This will be the case if we left the loop before it reached its end, i.e.
         before j becomes equal to nb_found_connected_sets. */
      // Assert.assertFalse (j == nb_found_connected_sets) ; TODO err msg
    }

    /* Now, connected-sets and pieces with their reference on these connected-sets
       are saved. The array of found connected-sets is no more useful, free it. */
    found_connected_sets = null ;
  }



  public static void saveGame (String file_name, Game game)
  	throws IOException
  {
    int tmp ;

    File sdCard = Environment.getExternalStorageDirectory () ; 
    File out_file = new File (sdCard, file_name) ;
    /* If a save game already exists, then silently kill it. */
    if (out_file.exists ()) out_file.delete () ;
    /* Now, create the save game file. */
    out_file.createNewFile () ;
    DataOutputStream out =
      new DataOutputStream (new FileOutputStream (out_file)) ;

    /* The magic number. */
    tmp = GameStruct.SAVEMAGICNUMBER ;
    out.writeInt (tmp) ;
    /* Miscellaneous scalars. */
    out.writeBoolean (game.rotated_image) ;
    out.writeInt (game.image_width) ;
    out.writeInt (game.image_height) ;
    out.writeFloat (game.original_image_scale_factor) ;
    out.writeInt (game.piece_width) ;
    out.writeInt (game.piece_height) ;
    out.writeInt (game.layout_nb_columns) ;
    out.writeInt (game.layout_nb_lines) ;
    out.writeInt (game.nb_pieces) ;

    /* The ids of the heads and tails of the x and y sorted lists of edges or -1
       if NULL. */
    if (game.x_sorted_edges_head == null) tmp = -1 ;
    else tmp = game.x_sorted_edges_head.id ;
    out.writeInt (tmp) ;
    if (game.x_sorted_edges_tail == null) tmp = -1 ;
    else tmp = game.x_sorted_edges_tail.id ;
    out.writeInt (tmp) ;
    if (game.y_sorted_edges_head == null) tmp = -1 ;
    else tmp = game.y_sorted_edges_head.id ;
    out.writeInt (tmp) ;
    if (game.y_sorted_edges_tail == null) tmp = -1 ;
    else tmp = game.y_sorted_edges_tail.id ;
    out.writeInt (tmp) ;
    
    /* Save pieces before, hence we will load them first and be able to create
       them to reference them by their id in the save of the layout. */
    savePieces (out, game.pieces, game.nb_pieces) ;

    saveLayout
    	(out, game.layout, (game.layout_nb_columns * game.layout_nb_lines)) ;

    /* The image file data. */
    out.writeInt (game.image_file_size) ;
    out.write (game.image_file_data, 0, game.image_file_size) ;
    
    out.close () ;
  }

}
