package francois.pessaux.androjig.android;

import android.graphics.Bitmap ;
import francois.pessaux.androjig.android.Maths2d.Fpoint ;
import francois.pessaux.androjig.android.Maths2d.Point ;
import francois.pessaux.androjig.android.SplineStruct.Spline ;

public class PieceStruct {

  /** Represents an edge of a piece of the jigsaw. Each piece own 4
      edges.
      Anchor points ids layout:
      0- 1- 2- 3- 4
      |  |  |  |  |
      5- 6- 7- 8-10
      |  |  |  |  |
      11-12-13-14-15 */
  public static class Edge {
    int id ;
    /** Boolean flag telling if the edge is still important and must be
        taken into account during computations. */
    boolean active ;
    /** Coordinate in the screen space of the first point of the edge. */
    Point p0 ;
    /** Unique id representing the first point of the edge. This point
        is an anchor point and must match another anchor point of another
        piece in order to have pieces coalescing. */
    int p0_anchor_id ;
    /* Coordinate in the screen space of the second point of the edge. */
    Point p1 ;
    /** Unique id representing the second point of the edge. This point
        is an anchor point and must match another anchor point of another
        piece in order to have pieces coalescing. */
    int p1_anchor_id ;
    /** The spline representing the shape of the piece. */
    Spline spline ;
    /** Pointer to the piece this edge belongs to. */
    Piece part_of_piece ;
    /** Pointer representing this edge's belonging to a double linked
        list where edges are sorted bu increasing minimal x
        coordinate. This pointer points to the next edge, having a
        minimal x coordinate greated or equal to the current edge's
        one. This list is used to find faster which edges are close to
        an edge by walking forward and backward in this list from the
        currently selected piece. */
    Edge x_sorted_prev ;
    /** Pointer representing this edge's belonging to a double linked
        list where edges are sorted by increasing minimal x
        coordinate. This pointer points to the previous edge, having a
        minimal x coordinate lower or equal to the current edge's
        one. This list is used to find faster which edges are close to
        an edge by walking forward and backward in this list from the
        currently selected piece. */
    Edge x_sorted_next ;
    /** Pointer representing this edge's belonging to a double linked
      list where edges are sorted by increasing minimal y
        coordinate. This pointer points to the next edge, having a
        minimal y coordinate greated or equal to the current edge's
        one. This list is used to find faster which edges are close to
        an edge by walking forward and backward in this list from the
        currently selected piece. */
    Edge y_sorted_prev ;
    /** Pointer representing this edge's belonging to a double linked
        list where edges are sorted bu increasing minimal y
        coordinate. This pointer points to the previous edge, having a
        minimal y coordinate lower or equal to the current edge's
        one. This list is used to find faster which edges are close to
        an edge by walking forward and backward in this list from the
        currently selected piece. */
    Edge y_sorted_next ;
  };



  public static final int NORTH = 0 ;
  public static final int EAST = 1 ;
  public static final int SOUTH = 2 ;
  public static final int WEST  = 3 ;

  
  
  /** Represents a 4-sided piece of the jigsaw. */
  public static class Piece {
    int piece_debug_id ;
    /** Position of the center of the piece in a graphic coordinates
        system, not in a mathematical one. Used to faster determine
        which piece is under the mouse pointer. */
    Point center ;
    /** Delta in x and y of the top left point of the piece's bounding
        box regarding to the center of the piece. */
    Point delta_center_to_top_left ;
    /** Delta in x and y of the bottom right point of the piece's bounding
        box regarding to the center of the piece. */
    Point delta_center_to_bottom_right ;
    /** Initial order N-E-S-O (indices 0-1-2-3) form a polygon. Once an
        edge matches another, hence 2 pieces coalesce, these 2 edges
        are destroyed from the pieces and removed from the double-linked
        lists to save computation. The edges array always contains 4 edges. */
    Edge[/* 4 */] free_edges ;
    /** Reference on the connected-set this piece belongs to if this piece
        is connected to other pieces. At least contains the piece itself. */
    Connected_set connected_pieces ;
    Bitmap piece_image ;
  };



  public static class Edges_list {
    Edge edge ;
    Edges_list next ;
  };



  public static class Pieces_list {
    Piece piece ;
    Pieces_list next ;
  };



  public static class Connected_set {
    /** The list of all the pieces connected together. */
    Pieces_list members ;
    /* Count the number of pieces structures pointing (sharing) this connected-set
       structure. This is used to prevent disallocation while some piece
       structures still point on the connected-set and also allows to free it
       only once.
       Moreover, this allows to know when the game ends: there must remain only
       one connected-set, hence the refcount must be equal to the number of pieces
       of the game. */
    int refcount ;
  };

  
  public static final int NB_UNIT_EDGE_POINTS = 10 ;
  
  
  
  /** Pattern of an horizontal piece's edge. Distorsion in y with head
      pointing toward the negative coordinates. The shape is normalized
      at the origine. */
  private static Fpoint[] unit_horizontal_edge ;


  /** Pattern of an horizontal piece's edge.  Distortion in x with the
      head pointing toward the negative coordinates. The shape is
  normalised at the origine. This is simply obtained by a rotation
  of -90 (math coords system) of the above previous shape (y' = -x
  and x' = y) and a shift "up" of 1.0 to have its "lower" point
  centered on the origin. Hence, such an edge is suitable to be the
  west edge of a piece (i.e. drawn bottom up in the screen
  coordinates system). */
  private static Fpoint[] unit_vertical_edge ;

    
  
  public static void initConstants ()
  {
    unit_horizontal_edge = new Fpoint [NB_UNIT_EDGE_POINTS] ;
    unit_horizontal_edge[0] = new Fpoint ((float) 0.000000, (float) 0.000000) ;  /* Start point. */
    unit_horizontal_edge[1] = new Fpoint ((float) 0.028409, (float) 0.000000) ;
    unit_horizontal_edge[2] = new Fpoint ((float) 0.500000, (float) 0.000000) ;
    unit_horizontal_edge[3] = new Fpoint ((float) 0.318182, (float) -0.136364) ;
    unit_horizontal_edge[4] = new Fpoint ((float) (-0.227273 + 0.5), (float) (-0.590909 + 0.2)) ;
    unit_horizontal_edge[5] = new Fpoint ((float) (1.227273 - 0.5), (float) (-0.590909 + 0.2)) ;
    unit_horizontal_edge[6] = new Fpoint ((float) 0.681818, (float) -0.136364) ;
    unit_horizontal_edge[7] = new Fpoint ((float) 0.500000, (float) 0.000000) ;
    unit_horizontal_edge[8] = new Fpoint ((float) 0.971591, (float) 0.000000) ;
    unit_horizontal_edge[9] = new Fpoint ((float) 1.000000, (float) 0.000000) ;    /* End point. */

    unit_vertical_edge = new Fpoint[NB_UNIT_EDGE_POINTS] ;
    unit_vertical_edge[0] = new Fpoint ((float) 0.000000, (float) (1.0 + 0.000000)) ;    /* Start point. */
    unit_vertical_edge[1] = new Fpoint ((float) 0.000000, (float) (1.0 + -0.028409)) ;
    unit_vertical_edge[2] = new Fpoint ((float) 0.000000, (float) (1.0 + -0.500000)) ;
    unit_vertical_edge[3] = new Fpoint ((float) -0.136364, (float) (1.0 + -0.318182)) ;
    unit_vertical_edge[4] = new Fpoint ((float) (-0.590909 + 0.2), (float) (1.0 + -(-0.227273 + 0.5))) ;
    unit_vertical_edge[5] = new Fpoint ((float) (-0.590909 + 0.2), (float) (1.0 + -(1.227273 - 0.5))) ;
    unit_vertical_edge[6] = new Fpoint ((float) -0.136364, (float) (1.0 + -0.681818)) ;
    unit_vertical_edge[7] = new Fpoint ((float) 0.000000, (float) (1.0 + -0.500000)) ;
    unit_vertical_edge[8] = new Fpoint ((float) 0.000000, (float) (1.0 + -0.971591)) ;
    unit_vertical_edge[9] = new Fpoint ((float) 0.000000, (float) (1.0 + -1.000000)) ;     /* End point. */
  }



  /** [Descr] : Creates the distortion of an horizontal edge. Basically,
      we have an horizontal pattern of noise of unit size and we scale
      it to the piece width. The distortion is the moved points through
      which we will create an approximating spline in order to create
      "little ears" on our edge.
      [Params] :
       - start_x : The x coordinate where the edge starts.
       - start_y : The y coordinate where the edge starts.
       - width : The length of the edge. Hence the edge will span from
         (start_x, start_y) to (start_x + width, start_y) bounds included.
      [Returns] : A pointer representing an array of length
         NB_UNIT_EDGE_POINTS containing some struct point.
      [Rem] : Exported outside this module. */
  public static Point[] makeHorizontalEdgeSegments (int start_x, int start_y,
                                                    int width)
  {
    float width_as_float = (float) width ;
    int y_flip = (AndroJigApp.randomGenerator.nextInt (2) == 0 ? 1 : -1) ;
    Point[] result_array ;
    int i ;

    result_array = new Point [NB_UNIT_EDGE_POINTS] ;

    for (i = 0; i < NB_UNIT_EDGE_POINTS; i++) {
      float fx = unit_horizontal_edge[i].fx ;
      float fy = unit_horizontal_edge[i].fy ;
      result_array[i] = new Point () ;
      
      if ((i == 4) || (i == 5)) {
        float px = /* +/- 0.2 to randomize the shape. */
          (float) (((float) (200 - AndroJigApp.randomGenerator.nextInt (400))) / 1000.0) ;
        result_array[i].x = ((int) ((fx + px) * width_as_float)) + start_x ;
        result_array[i].y = y_flip * ((int) (fy * width_as_float)) + start_y ;
      }
      else {
        result_array[i].x = ((int) (fx * width_as_float)) + start_x ;
        result_array[i].y = y_flip * ((int) (fy * width_as_float)) + start_y ;
      }
    }        /* End of for (i = 0; i < NB_UNIT_EDGE_POINTS; i++). */

    return (result_array) ;
  }



  /** [Descr] : Creates the distortion of a vertical edge. Basically,
      we have an vertical pattern of noise of unit size and we scale
      it to the piece height. The distortion is the moved points through
      which we will create an approximating spline in order to create
      "little ears" on our edge.
      [Params] :
       - start_x : The x coordinate where the edge starts.
       - start_y : The y coordinate where the edge starts.
       - height : The height of the edge. Hence the edge will span from
         (start_x, start_y) to (start_x, start_y + height) bounds included.
      [Rem] : Exported outside this module. */
  public static Point[] makeVerticalEdgeSegments (int start_x, int start_y,
                                                  int height)
  {
    float height_as_float = (float) height ;
    int x_flip = (AndroJigApp.randomGenerator.nextInt (2) == 0 ? 1 : -1) ;
    Point[] result_array ;
    int i ;

    result_array = new Point [NB_UNIT_EDGE_POINTS] ;

    for (i = 0; i < NB_UNIT_EDGE_POINTS; i++) {
      float fx = unit_vertical_edge[i].fx ;
      float fy = unit_vertical_edge[i].fy ;
      result_array[i] = new Point () ;
      
      if ((i == 4) || (i == 5)) {
        float py = /* +/- 0.2 to randomize the shape. */
          (float) (((float) (200 - AndroJigApp.randomGenerator.nextInt (400))) / 1000.0) ;
        result_array[i].x = (x_flip * ((int) (fx * height_as_float))) + start_x ;
        result_array[i].y = ((int) ((fy + py) * height_as_float)) + start_y ;
      }
      else {
        result_array[i].x = x_flip * ((int) (fx * height_as_float)) + start_x ;
        result_array[i].y = ((int) (fy * height_as_float)) + start_y ;
      }
    }

return (result_array) ;
}



  /** Releases memory occupied by a list of pieces, but not the pieces
      themselves. */
  public static void freePiecesList (Pieces_list lst)
  {
    Pieces_list tmp ;

    while (lst != null) {
      tmp = lst.next ;
      lst.next = null ;
      lst = tmp ;
    }
}



public static Pieces_list copyPiecesList  (Pieces_list lst)
{
Pieces_list copy_head ;
Pieces_list tmp ;
Pieces_list prev ;

copy_head = null ;
prev = null ;

while (lst != null) {
  tmp = new Pieces_list () ;
  tmp.piece = lst.piece ;
  tmp.next = null ;
  if (copy_head == null) copy_head = tmp ;
  if (prev != null) prev.next = tmp ;
  prev = tmp ;
  lst = lst.next ;
}

return (copy_head) ;
}



/** Releases memory occupied by a list of edges, but not the edges
  themselves. */
public static void freeEdgesList (Edges_list lst)
{
Edges_list tmp ;

while (lst != null) {
  tmp = lst.next ;
  lst.next = null ;
  lst = tmp ;
}
}


public static int piecesListLength (Pieces_list list)
{
int len = 0 ;
while (list != null) {
  len++ ;
  list = list.next ;
}

return (len) ;
}

}
