package francois.pessaux.androjig.android;

import java.util.Arrays ;

import android.graphics.Bitmap ;
import francois.pessaux.androjig.android.Maths2d.Point ;
import francois.pessaux.androjig.android.PieceStruct.Piece ;

public class TextureMap {

  public static class Segment {
    Point p0 ;
    Point p1 ;
  };
  
  
  private static boolean addMaskToBitmap (int[] pixels,
      int dest_width, int dest_height, Segment[] segments, int nbseg)
  {
    int i, j, k, yh, left, right, cpt, aux ;
    int[] indir ;
    float[] invp ;
    float[] intersec ;

    indir = new int [nbseg] ;
    invp = new float [nbseg] ;
    intersec = new float [nbseg] ;

    /* Indirection table allowing to sort edges by increasing y initially
       initialized with identity. */
    for (i = 0; i < nbseg; i++) indir[i] = i ;

    /* Make sure that for each edge, the point with the minimal y is the
       first one (the start point). */
    for (i = 0; i <nbseg; i++) {
      if (segments[i].p0.y > segments[i].p1.y) {
        int x_max = segments[i].p0.x ;
        int y_max = segments[i].p0.y ;
        int x_min = segments[i].p1.x ;
        int y_min = segments[i].p1.y ;
        segments[i].p0.x = x_min ;
        segments[i].p0.y = y_min ;
        segments[i].p1.x = x_max ;
        segments[i].p1.y = y_max ;
      }
    }
    /* Remove horizontal edges and shift remaining edges to the left
       in the indirection array (do not modify the edges themselves). */
    i = 0 ;
    while (i <= (nbseg - 1)) {
      if (segments[indir[i]].p0.y == segments[indir[i]].p1.y) {
        for (j = i + 1; j < nbseg; j++) {
          indir[j - 1] = indir[j] ;
        }
        nbseg-- ;
      }
      else i++ ;
    }
    /* Sort edges in the indirection table so that y are in increasing
       order. */
    for (k = nbseg - 1; k >= 1; k--) {
      int indmax = 0 ;
      for (j = 1; j <= k; j++) {
        if (segments[indir[indmax]].p0.y < segments[indir[j]].p0.y) {
          indmax = j ;
        }
        aux = indir[indmax] ;
        indir[indmax] = indir[k] ;
        indir[k] = aux ;
      }    /* End of for (j = 1; j <= k; j++). */
    }      /* End of for (k = nbseg - 1; k >= 1; k--). */
    /* */
    for (i = 0; i < nbseg; i++) {
      invp[indir[i]] =
        ((float) (segments[indir[i]].p1.x - segments[indir[i]].p0.x))
        /
        ((float) (segments[indir[i]].p1.y - segments[indir[i]].p0.y)) ;
    }
    /* */
    yh = segments[indir[0]].p0.y ;
    right = 0 ;
    left = -1 ;
    while (left < (nbseg - 1)) {
      while ((segments[indir[right]].p0.y <= yh) &&
          (right <= (nbseg - 1))) {
        right = right + 1 ;
      }
      for (k = left + 1; k <= (right - 1); k++) {
        if (segments[indir[k]].p1.y <= yh) {
          left = left + 1 ;
          aux = indir[left] ;
          indir[left] = indir[k] ;
          indir[k] = aux ;
        }
      }
      /* */
      cpt = 0 ;
      for (k = left + 1; k <= (right - 1); k++) {
        intersec[cpt] =
          ((float) segments[indir[k]].p0.x) +
          ((float) (yh - segments[indir[k]].p0.y)) *
          invp[indir[k]] ;
        cpt++ ;
      }
      if (cpt > 0) {
        int line, x ;

        /* Sort the array on the cpt first elements, in increasing order. */
        Arrays.sort (intersec, 0, cpt) ;

        line = 0 ;
        while (line <= (cpt - 1)) {
          int x_start = (int) intersec[line] ;
          int x_stop = (int) intersec[line + 1] ;
          /* Factorise the multiplication by the number of pixels due to vertical offset. */
          int pixel_index = dest_width * yh + x_start ;          
          for (x = x_start; x <= x_stop; x++) {
            if ((yh >= 0) && (yh < dest_height) &&
                (x >= 0) && (x < dest_width)) {
              pixels[pixel_index] = pixels[pixel_index] | 0xFF000000 ;
            }
            /* Simply increment offset to access next pixel from the same line. */
            pixel_index++ ;
          }
          line += 2 ;
        }
      }
      yh++ ; /* Next line. */
    }      /* End of while (left < (nbseg - 1)). */

    indir = null ;
    invp = null ;
    intersec = null ;
    return (true) ;
  }



  /** Function used to create the piece texture with transparent areas. It maps
      the related original image part of the picture to the piece and makes
      transparent area outside the interior of the polygon described by the
      splines of the piece's edges. */
  public static boolean initializePieceTexture (Bitmap original_image,
                                                Piece piece,
                                                Point piece_initial_center)
  {
    int i, j, count ;
    int x_shift, y_shift ;
    int dest_img_width, dest_img_height ;
    int x_source_offset, y_source_offset ;
    Segment[] segments ;

    /* Count the total number of edges in all the splines of the
       piece. */
    int nb_edges = 0 ;
    for (i = 0; i < 4; i++) {
      /* Attention, the last point doesn't count for an edge. Problem
         of posts and intervals... */
      nb_edges += (piece.free_edges[i].spline.nb_points - 1) ;
    }

    /* Create the array of segments to store the polygon description
       issued from the catenation of all the edges of the piece. We
       initialize the points coordinates with a default value and we
       will update it later. */
    segments = new Segment [nb_edges] ;
    for (j = 0; j < nb_edges; j++) {
      segments[j] = new Segment () ;
      segments[j].p0 = new Point () ;
      segments[j].p1 = new Point () ;
    }

    /* Compute the delta in x and y to substract to each segment in
       order to simulate the fact that the piece is at (0, 0). Hence,
       filling routine will fill in the piece's image which is a simple
       matrix starting at (0, 0) and ending at (piece_width, piece_height)
       with boundaries included. */
    x_shift = piece_initial_center.x + piece.delta_center_to_top_left.x ;
    y_shift = piece_initial_center.y + piece.delta_center_to_top_left.y ;
    /* Update the segments coordinates from the splines points
       coordinates. */
    count = 0 ;
    for (j = 0;  j < 4; j++) {
      PieceStruct.Edge edge = piece.free_edges[j] ;
      for (i = 0; i < edge.spline.nb_points - 1; i++) {
        segments[count].p0.x = edge.spline.vertices[i].x - x_shift ;
        segments[count].p0.y = edge.spline.vertices[i].y - y_shift ;
        segments[count].p1.x = edge.spline.vertices[i + 1].x - x_shift ;
        segments[count].p1.y = edge.spline.vertices[i + 1].y - y_shift ;
        count++ ;
      }
    }     /* End of for (j = 0;  j < 4; j++). */

    /* Compute the with and height of the area to map, i.e. the piece's extend. */
    dest_img_height =
      piece.delta_center_to_bottom_right.y -
      piece.delta_center_to_top_left.y ;
    dest_img_width =
      piece.delta_center_to_bottom_right.x -
      piece.delta_center_to_top_left.x ;

    /* We need the x and y offsets in the source image where to find colors to put
       in the destination image. In effect, the destination image space spans from
       (0, 0) to (piece_width, piece_height) but its points' colors must be taken
       in the source image at the place the piece lays. */
    x_source_offset =
      piece_initial_center.x + piece.delta_center_to_top_left.x ;
    y_source_offset =
      piece_initial_center.y + piece.delta_center_to_top_left.y ;

    /* Set transparency initially on all the pixels of the piece image. */
    int [] pixels = new int [dest_img_width * dest_img_height] ;
    original_image.getPixels
      (pixels, 0, dest_img_width, x_source_offset, y_source_offset,
       dest_img_width, dest_img_height) ;
    for (i = 0; i < dest_img_width * dest_img_height; i++)
      pixels[i] = pixels[i] & 0x00FFFFFF ;
    
    /* Mask the image of the piece following the splines shape. */
    addMaskToBitmap (pixels, dest_img_width, dest_img_height, segments, nb_edges) ;

    Bitmap sub_image =
        Bitmap.createBitmap
          (dest_img_width, dest_img_height, Bitmap.Config.ARGB_8888) ;
    sub_image.setPixels
      (pixels, 0, dest_img_width, 0, 0, dest_img_width, dest_img_height) ;

    /* Release pointer on the bytes since they are no more useful. */
    pixels = null ;
    /* Release the segments since they are no more useful. */
    segments = null ;

    /* Finally, set the image information in the piece's structure. */
    piece.piece_image = sub_image ;
    return (true) ;
  }

}
